import time
import os
import yaml
import httpx
from typing import Any, Dict, Tuple, Type, Union, Optional

from langchain_core.vectorstores import VectorStoreRetriever
from loguru import logger
from langchain_core.tools import BaseTool, Tool
from langchain_core.pydantic_v1 import BaseModel, Extra, Field, root_validator
from langchain_core.language_models.base import LanguageModelLike
from langchain_core.prompts import ChatPromptTemplate
from langchain.chains.llm import LLMChain
from langchain.chains.question_answering import load_qa_chain
from bisheng_langchain.retrievers import EnsembleRetriever
from bisheng_langchain.vectorstores import ElasticKeywordsSearch, Milvus
from bisheng_langchain.rag.init_retrievers import (
    BaselineVectorRetriever,
    KeywordRetriever,
    MixRetriever,
    SmallerChunksVectorRetriever,
)
from bisheng_langchain.rag.utils import import_by_type, import_class
from bisheng_langchain.rag.extract_info import extract_title


class MultArgsSchemaTool(Tool):

    def _to_args_and_kwargs(self, tool_input: Union[str, Dict]) -> Tuple[Tuple, Dict]:
        # For backwards compatibility, if run_input is a string,
        # pass as a positional argument.
        if isinstance(tool_input, str):
            return (tool_input, ), {}
        else:
            return (), tool_input


class BishengRAGTool:

    def __init__(
        self,
        vector_store: Optional[Milvus] = None,
        keyword_store: Optional[ElasticKeywordsSearch] = None,
        llm: Optional[LanguageModelLike] = None,
        collection_name: Optional[str] = None,
        QA_PROMPT: Optional[ChatPromptTemplate] = None,
        **kwargs
    ) -> None:
        if collection_name is None and (keyword_store is None or vector_store is None):
            raise ValueError('collection_name must be provided if keyword_store or vector_store is not provided')
        self.collection_name = collection_name
        
        yaml_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'config/baseline_v2.yaml')
        with open(yaml_path, 'r', encoding='utf-8') as f:
            self.params = yaml.safe_load(f)
        
        # update params
        max_content = kwargs.get("max_content", 15000)
        sort_by_source_and_index = kwargs.get("sort_by_source_and_index", True)
        self.params['generate']['max_content'] = max_content
        self.params['post_retrieval']['sort_by_source_and_index'] = sort_by_source_and_index

        # init llm
        if llm:
            self.llm = llm
        else:
            llm_params = self.params['chat_llm']
            llm_object = import_by_type(_type='llms', name=llm_params['type'])
            if llm_params['type'] == 'ChatOpenAI' and llm_params['openai_proxy']:
                llm_params.pop('type')
                self.llm = llm_object(http_client=httpx.Client(proxies=llm_params['openai_proxy']), **llm_params)
            else:
                llm_params.pop('type')
                self.llm = llm_object(**llm_params)
        
        # init milvus
        if vector_store:
            # if vector_store is retriever, get vector_store instance
            if isinstance(vector_store, VectorStoreRetriever):
                self.vector_store = vector_store.vectorstore
            else:
                self.vector_store = vector_store
        else:
            # init embeddings
            embedding_params = self.params['embedding']
            embedding_object = import_by_type(_type='embeddings', name=embedding_params['type'])
            if embedding_params['type'] == 'OpenAIEmbeddings' and embedding_params['openai_proxy']:
                embedding_params.pop('type')
                self.embeddings = embedding_object(
                    http_client=httpx.Client(proxies=embedding_params['openai_proxy']), **embedding_params
                )
            else:
                embedding_params.pop('type')
                self.embeddings = embedding_object(**embedding_params)
            
            self.vector_store = Milvus(
                embedding_function=self.embeddings,
                connection_args={
                    "host": self.params['milvus']['host'],
                    "port": self.params['milvus']['port'],
                },
            )
        
        # init keyword store
        if keyword_store:
            self.keyword_store = keyword_store
        else:
            if self.params['elasticsearch'].get('extract_key_by_llm', False):
                extract_key_prompt = import_class(f'bisheng_langchain.rag.prompts.EXTRACT_KEY_PROMPT')
                llm_chain = LLMChain(llm=self.llm, prompt=extract_key_prompt)
            else:
                llm_chain = None
            self.keyword_store = ElasticKeywordsSearch(
                index_name='default_es',
                elasticsearch_url=self.params['elasticsearch']['url'],
                ssl_verify=self.params['elasticsearch']['ssl_verify'],
                llm_chain=llm_chain,
            )

        # init retriever
        retriever_list = []
        retrievers = self.params['retriever']['retrievers']
        for retriever in retrievers:
            retriever_type = retriever.pop('type')
            retriever_params = {
                'vector_store': self.vector_store,
                'keyword_store': self.keyword_store,
                'splitter_kwargs': retriever['splitter'],
                'retrieval_kwargs': retriever['retrieval'],
            }
            retriever_list.append(self._post_init_retriever(retriever_type=retriever_type, **retriever_params))
        self.retriever = EnsembleRetriever(retrievers=retriever_list)

        # init qa chain    
        if QA_PROMPT:
            prompt = QA_PROMPT
        else:
            if 'prompt_type' in self.params['generate']:
                prompt_type = self.params['generate']['prompt_type']
                prompt = import_class(f'bisheng_langchain.rag.prompts.{prompt_type}')
            else:
                prompt = None
        self.qa_chain = load_qa_chain(
            llm=self.llm, 
            chain_type=self.params['generate']['chain_type'], 
            prompt=prompt, 
            verbose=False
        )
    
    def _post_init_retriever(self, retriever_type, **kwargs):
        retriever_classes = {
            'KeywordRetriever': KeywordRetriever,
            'BaselineVectorRetriever': BaselineVectorRetriever,
            'MixRetriever': MixRetriever,
            'SmallerChunksVectorRetriever': SmallerChunksVectorRetriever,
        }
        if retriever_type not in retriever_classes:
            raise ValueError(f'Unknown retriever type: {retriever_type}')

        input_kwargs = {}
        splitter_params = kwargs.pop('splitter_kwargs')
        for key, value in splitter_params.items():
            splitter_obj = import_by_type(_type='textsplitters', name=value.pop('type'))
            input_kwargs[key] = splitter_obj(**value)

        retrieval_params = kwargs.pop('retrieval_kwargs')
        for key, value in retrieval_params.items():
            input_kwargs[key] = value

        input_kwargs['vector_store'] = kwargs.pop('vector_store')
        input_kwargs['keyword_store'] = kwargs.pop('keyword_store')

        retriever_class = retriever_classes[retriever_type]
        return retriever_class(**input_kwargs)

    def file2knowledge(self, file_path, drop_old=True):
        """
        file to knowledge
        """
        loader_params = self.params['loader']
        loader_object = import_by_type(_type='documentloaders', name=loader_params.pop('type'))

        logger.info(f'file_path: {file_path}')
        loader = loader_object(
            file_name=os.path.basename(file_path), file_path=file_path, **loader_params
        )
        documents = loader.load()
        logger.info(f'documents: {len(documents)}, page_content: {len(documents[0].page_content)}')
        if len(documents[0].page_content) == 0:
            logger.error(f'{file_path} page_content is empty.')

        # add aux info
        add_aux_info = self.params['retriever'].get('add_aux_info', False)
        if add_aux_info:
            for doc in documents:
                try:
                    title = extract_title(llm=self.llm, text=doc.page_content)
                    logger.info(f'extract title: {title}')
                except Exception as e:
                    logger.error(f"Failed to extract title: {e}")
                    title = ''
                doc.metadata['title'] = title

        for idx, retriever in enumerate(self.retriever.retrievers):
            retriever.add_documents(
                documents, 
                self.collection_name, 
                drop_old=drop_old, 
                add_aux_info=add_aux_info
            )
    
    def retrieval_and_rerank(self, query):
        """
        retrieval and rerank
        """
        # EnsembleRetriever直接检索召回会默认去重
        docs = self.retriever.get_relevant_documents(
            query=query, 
            collection_name=self.collection_name
        )
        logger.info(f'retrieval docs origin: {len(docs)}')

        # delete redundancy according to max_content 
        doc_num, doc_content_sum = 0, 0
        for doc in docs:
            doc_content_sum += len(doc.page_content)
            if doc_content_sum > self.params['generate']['max_content']:
                break
            doc_num += 1
        docs = docs[:doc_num]
        logger.info(f'retrieval docs after delete redundancy: {len(docs)}')

        # 按照文档的source和chunk_index排序，保证上下文的连贯性和一致性
        if self.params['post_retrieval'].get('sort_by_source_and_index', False):
            logger.info('sort chunks by source and chunk_index')
            docs = sorted(docs, key=lambda x: (x.metadata['source'], x.metadata['chunk_index']))
        return docs
    
    def run(self, query, return_only_outputs=True) -> Any:
        docs = self.retrieval_and_rerank(query)
        try:
            ans = self.qa_chain({"input_documents": docs, "question": query}, return_only_outputs=return_only_outputs)
        except Exception as e:
            logger.error(f'question: {query}\nerror: {e}')
            ans = {'output_text': str(e)}
        if return_only_outputs:
            rag_answer = ans['output_text']
            return rag_answer
        else:
            rag_answer = ans['output_text']
            input_documents = ans['input_documents']
            return rag_answer, input_documents
    
    async def arun(self, query: str, return_only_outputs=True) -> str:
        rag_answer = self.run(query, return_only_outputs)
        return rag_answer
    
    @classmethod
    def get_rag_tool(cls, name, description, **kwargs: Any) -> BaseTool:
        class InputArgs(BaseModel):
            """Agent will use for structed output generating under ReAct mode"""
            query: str = Field(description='question asked by the user.')
            class Config:
                schema_extra = {
                    "properties": {'query': {'type': 'string'}}
                    }

        return MultArgsSchemaTool(name=name,
                                  description=description,
                                  func=cls(**kwargs).run,
                                  coroutine=cls(**kwargs).arun,
                                  args_schema=InputArgs)
    

if __name__ == '__main__':
    # rag_tool = BishengRAGTool(collection_name='rag_finance_report_0_test')
    # rag_tool.file2knowledge(file_path='/home/public/rag_benchmark_finance_report/金融年报财报的来源文件/2021-04-23__金宇生物技术股份有限公司__600201__生物股份__2020年__年度报告.pdf')

    from langchain.chat_models import ChatOpenAI
    from langchain.embeddings import OpenAIEmbeddings
    # embedding
    embeddings = OpenAIEmbeddings(model='text-embedding-ada-002')
    # llm
    llm = ChatOpenAI(model='gpt-4-1106-preview', temperature=0.01)
    collection_name = 'rag_finance_report_0_benchmark_caibao_1000_source_title'
    # milvus
    vector_store = Milvus(
            collection_name=collection_name,
            embedding_function=embeddings,
            connection_args={
                "host": '110.16.193.170',
                "port": '50062',
            },
    )
    # es
    keyword_store = ElasticKeywordsSearch(
        index_name=collection_name,
        elasticsearch_url='http://110.16.193.170:50062/es',
        ssl_verify={'basic_auth': ["elastic", "oSGL-zVvZ5P3Tm7qkDLC"]},
    )

    tool = BishengRAGTool.get_rag_tool(
        name='rag_knowledge_retrieve', 
        description='金融年报财报知识库问答',
        vector_store=vector_store, 
        keyword_store=keyword_store, 
        llm=llm
    )
    print(tool.run('能否根据2020年金宇生物技术股份有限公司的年报，给我简要介绍一下报告期内公司的社会责任工作情况？'))

    # tool = BishengRAGTool.get_rag_tool(
    #     name='rag_knowledge_retrieve', 
    #     description='金融年报财报知识库问答',
    #     collection_name='rag_finance_report_0_benchmark_caibao_1000_source_title'
    # )
    # print(tool.run('能否根据2020年金宇生物技术股份有限公司的年报，给我简要介绍一下报告期内公司的社会责任工作情况？'))


from __future__ import annotations

import json
import logging
import time
from typing import Any, Dict, List, Mapping, Optional, Tuple, Union

import jwt
from bisheng_langchain.utils.requests import Requests
from langchain.callbacks.manager import AsyncCallbackManagerForLLMRun, CallbackManagerForLLMRun
from langchain.chat_models.base import BaseChatModel
from langchain.schema import ChatGeneration, ChatResult
from langchain.schema.messages import (AIMessage, BaseMessage, ChatMessage, FunctionMessage,
                                       HumanMessage, SystemMessage)
from langchain.utils import get_from_dict_or_env
from langchain_core.pydantic_v1 import Field, root_validator
from tenacity import (before_sleep_log, retry, retry_if_exception_type, stop_after_attempt,
                      wait_exponential)

# if TYPE_CHECKING:
#     import jwt

logger = logging.getLogger(__name__)


def _import_pyjwt() -> Any:
    try:
        import jwt
    except ImportError:
        raise ValueError('Could not import jwt python package. '
                         'This is needed in order to calculate get_token_ids. '
                         'Please install it with `pip install PyJWT`.')
    return jwt


def encode_jwt_token(ak, sk):
    headers = {'alg': 'HS256', 'typ': 'JWT'}
    payload = {
        'iss': ak,
        'exp': int(time.time()) + 18000,  # 填写您期望的有效时间，此处示例代表当前时间+300分钟
        'nbf': int(time.time()) - 500  # 填写您期望的生效时间，此处示例代表当前时间-500秒
    }
    token = jwt.encode(payload, sk, headers=headers)
    return token


def _create_retry_decorator(llm):

    min_seconds = 1
    max_seconds = 20
    # Wait 2^x * 1 second between each retry starting with
    # 4 seconds, then up to 10 seconds, then 10 seconds afterwards
    return retry(
        reraise=True,
        stop=stop_after_attempt(llm.max_retries),
        wait=wait_exponential(multiplier=1, min=min_seconds, max=max_seconds),
        retry=(retry_if_exception_type(Exception)),
        before_sleep=before_sleep_log(logger, logging.WARNING),
    )


def _convert_dict_to_message(_dict: Mapping[str, Any]) -> BaseMessage:
    role = _dict['role']
    if role == 'user':
        return HumanMessage(content=_dict['content'])
    elif role == 'assistant':
        content = _dict['content'] or ''  # OpenAI returns None for tool invocations

        if _dict.get('function_call'):
            additional_kwargs = {'function_call': dict(_dict['function_call'])}
        else:
            additional_kwargs = {}
        return AIMessage(content=content, additional_kwargs=additional_kwargs)
    elif role == 'system':
        return SystemMessage(content=_dict['content'])
    elif role == 'function':
        return FunctionMessage(content=_dict['content'], name=_dict['name'])
    else:
        return ChatMessage(content=_dict['content'], role=role)


def _convert_message_to_dict(message: BaseMessage) -> dict:
    if isinstance(message, ChatMessage):
        message_dict = {'role': message.role, 'content': message.content}
    elif isinstance(message, HumanMessage):
        message_dict = {'role': 'user', 'content': message.content}
    elif isinstance(message, AIMessage):
        message_dict = {'role': 'assistant', 'content': message.content}
    elif isinstance(message, SystemMessage):
        message_dict = {'role': 'system', 'content': message.content}
        # raise ValueError(f"not support system role {message}")

    elif isinstance(message, FunctionMessage):
        raise ValueError(f'not support funciton {message}')
    else:
        raise ValueError(f'Got unknown type {message}')

    # if "name" in message.additional_kwargs:
    #     message_dict["name"] = message.additional_kwargs["name"]
    return message_dict


def _convert_message_to_dict2(message: BaseMessage) -> List[dict]:
    if isinstance(message, ChatMessage):
        message_dict = {'role': message.role, 'content': message.content}
    elif isinstance(message, HumanMessage):
        message_dict = {'role': 'user', 'content': message.content}
    elif isinstance(message, AIMessage):
        message_dict = {'role': 'assistant', 'content': message.content}
    elif isinstance(message, SystemMessage):
        raise ValueError(f'not support system role {message}')

    elif isinstance(message, FunctionMessage):
        raise ValueError(f'not support funciton {message}')
    else:
        raise ValueError(f'Got unknown type {message}')

    return [message_dict]


url = 'https://api.sensenova.cn/v1/llm/chat-completions'


class SenseChat(BaseChatModel):

    client: Optional[Any]  #: :meta private:
    model_name: str = Field(default='SenseChat', alias='model')
    """Model name to use."""
    temperature: float = 0.8
    top_p: float = 0.7
    """What sampling temperature to use."""
    model_kwargs: Optional[Dict[str, Any]] = Field(default_factory=dict)
    """Holds any model parameters valid for `create` call not explicitly specified."""
    access_key_id: Optional[str] = None
    secret_access_key: Optional[str] = None

    repetition_penalty: float = 1.05
    request_timeout: Optional[Union[float, Tuple[float, float]]] = None
    """Timeout for requests to OpenAI completion API. Default is 600 seconds."""
    max_retries: Optional[int] = 6
    """Maximum number of retries to make when generating."""
    streaming: Optional[bool] = False
    """Whether to stream the results or not."""
    n: Optional[int] = 1
    """Number of chat completions to generate for each prompt."""
    max_tokens: Optional[int] = 1024
    """Maximum number of tokens to generate."""
    tiktoken_model_name: Optional[str] = None
    """The model name to pass to tiktoken when using this class.
    Tiktoken is used to count the number of tokens in documents to constrain
    them to be under a certain limit. By default, when set to None, this will
    be the same as the embedding model name. However, there are some cases
    where you may want to use this Embedding class with a model name not
    supported by tiktoken. This can include when using Azure embeddings or
    when using one of the many model providers that expose an OpenAI-like
    API but with different models. In those cases, in order to avoid erroring
    when tiktoken is called, you can specify a model name to use here."""
    verbose: Optional[bool] = False

    class Config:
        """Configuration for this pydantic object."""

        allow_population_by_field_name = True

    @root_validator()
    def validate_environment(cls, values: Dict) -> Dict:
        """Validate that api key and python package exists in environment."""

        _import_pyjwt()

        values['access_key_id'] = get_from_dict_or_env(values, 'access_key_id', 'ACCESS_KEY_ID')
        values['secret_access_key'] = get_from_dict_or_env(values, 'secret_access_key',
                                                           'SECRET_ACCESS_KEY')
        token = encode_jwt_token(values['access_key_id'], values['secret_access_key'])
        if isinstance(token, bytes):
            token = token.decode('utf-8')

        try:
            header = {
                'Authorization': 'Bearer {}'.format(token),
                'Content-Type': 'application/json'
            }

            values['client'] = Requests(headers=header, )
        except AttributeError:
            raise ValueError('Try upgrading it with `pip install --upgrade requests`.')
        return values

    @property
    def _default_params(self) -> Dict[str, Any]:
        """Get the default parameters for calling ZhipuAI API."""
        return {
            'model': self.model_name,
            'temperature': self.temperature,
            'top_p': self.top_p,
            'max_tokens': self.max_tokens,
            **self.model_kwargs,
        }

    def completion_with_retry(self, **kwargs: Any) -> Any:
        retry_decorator = _create_retry_decorator(self)

        @retry_decorator
        def _completion_with_retry(**kwargs: Any) -> Any:
            messages = kwargs.get('messages')
            temperature = kwargs.get('temperature')
            top_p = kwargs.get('top_p')
            # messages
            params = {
                'messages': messages,
                'model': self.model_name,
                'top_p': top_p,
                'temperature': temperature,
                'repetition_penalty': self.repetition_penalty,
                'n': self.n,
                'max_new_tokens': self.max_tokens,
                'stream': False  # self.streaming
            }

            token = encode_jwt_token(self.access_key_id, self.secret_access_key)
            if isinstance(token, bytes):
                token = token.decode('utf-8')
            self.client.headers.update({'Authorization': 'Bearer {}'.format(token)})

            response = self.client.post(url=url, json=params).json()
            return response

        rsp_dict = _completion_with_retry(**kwargs)
        if 'error' in rsp_dict:
            logger.error(f'sensechat_error resp={rsp_dict}')
            message = rsp_dict['error']['message']
            raise Exception(message)
        else:
            # return rsp_dict['data'], rsp_dict.get('usage', '')
            return rsp_dict, rsp_dict.get('usage', '')

    async def acompletion_with_retry(self, **kwargs: Any) -> Any:
        """Use tenacity to retry the async completion call."""
        retry_decorator = _create_retry_decorator(self)

        if self.streaming:
            self.client.headers.update({'Accept': 'text/event-stream'})
        else:
            self.client.headers.pop('Accept', '')

        @retry_decorator
        async def _acompletion_with_retry(**kwargs: Any) -> Any:
            messages = kwargs.pop('messages', '')

            inp = {
                'messages': messages,
                'model': self.model_name,
                'top_p': self.top_p,
                'temperature': self.temperature,
                'repetition_penalty': self.repetition_penalty,
                'n': self.n,
                'max_new_tokens': self.max_tokens,
                'stream': True
            }

            # Use OpenAI's async api https://github.com/openai/openai-python#async-api
            async with self.client.apost(url=url, json=inp) as response:

                async for line in response.content.iter_any():

                    if b'\n' in line:
                        for txt_ in line.split(b'\n'):
                            yield txt_.decode('utf-8').strip()
                    else:
                        yield line.decode('utf-8').strip()

        async for response in _acompletion_with_retry(**kwargs):
            is_error = False
            if response:
                if response.startswith('event:error'):
                    is_error = True
                elif response.startswith('data:'):
                    yield (is_error, response[len('data:'):])
                    if is_error:
                        break
                elif response.startswith('{'):
                    yield (is_error, response)
                else:
                    continue

    def _combine_llm_outputs(self, llm_outputs: List[Optional[dict]]) -> dict:
        overall_token_usage: dict = {}
        for output in llm_outputs:
            if output is None:
                # Happens in streaming
                continue
            token_usage = output['token_usage']
            for k, v in token_usage.items():
                if k in overall_token_usage:
                    overall_token_usage[k] += v
                else:
                    overall_token_usage[k] = v
        return {'token_usage': overall_token_usage, 'model_name': self.model_name}

    def _generate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> ChatResult:

        message_dicts, params = self._create_message_dicts(messages, stop)
        params = {**params, **kwargs}
        response, usage = self.completion_with_retry(messages=message_dicts, **params)

        return self._create_chat_result(response)

    async def _agenerate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> ChatResult:
        message_dicts, params = self._create_message_dicts(messages, stop)
        params = {**params, **kwargs}
        if self.streaming:

            inner_completion = ''
            role = 'user'
            params['stream'] = True
            function_call: Optional[dict] = None
            async for is_error, stream_resp in self.acompletion_with_retry(messages=message_dicts,
                                                                           **params):
                if str(stream_resp).startswith('[DONE]'):
                    continue
                output = json.loads(stream_resp)
                if is_error:
                    logger.error(stream_resp)
                    raise ValueError(stream_resp)
                if 'data' in output:
                    output = output['data']

                choices = None
                if 'choices' in output:
                    choices = output.get('choices')

                if choices:
                    for choice in choices:
                        token = choice['delta']

                        inner_completion += token or ''
                        _function_call = ''
                        if run_manager:
                            await run_manager.on_llm_new_token(token)
                        if _function_call:
                            if function_call is None:
                                function_call = _function_call
                            else:
                                function_call['arguments'] += _function_call['arguments']
            message = _convert_dict_to_message({
                'content': inner_completion,
                'role': role,
                'function_call': function_call,
            })
            return ChatResult(generations=[ChatGeneration(message=message)])
        else:
            return self._generate(messages, stop, run_manager, **kwargs)

    def _create_message_dicts(
            self, messages: List[BaseMessage],
            stop: Optional[List[str]]) -> Tuple[List[Dict[str, Any]], Dict[str, Any]]:
        params = dict(self._client_params)
        if stop is not None:
            if 'stop' in params:
                raise ValueError('`stop` found in both the input and default params.')
            params['stop'] = stop

        system_content = ''
        message_dicts = []
        for m in messages:
            if m.type == 'system':
                system_content += m.content
                continue
            message_dicts.extend(_convert_message_to_dict2(m))

        if system_content:
            message_dicts[-1]['content'] = system_content + message_dicts[-1]['content']

        return message_dicts, params

    def _create_chat_result(self, response: Mapping[str, Any]) -> ChatResult:
        generations = []

        def _norm_text(text):
            if text[0] == '"' and text[-1] == '"':
                out = eval(text)
            else:
                out = text
            return out

        for res in response['data']['choices']:
            res['content'] = _norm_text(res['message'])
            res['role'] = 'user'
            message = _convert_dict_to_message(res)
            gen = ChatGeneration(message=message)
            generations.append(gen)

        llm_output = {'token_usage': response['data']['usage'], 'model_name': self.model_name}
        return ChatResult(generations=generations, llm_output=llm_output)

    @property
    def _identifying_params(self) -> Mapping[str, Any]:
        """Get the identifying parameters."""
        return {**{'model_name': self.model_name}, **self._default_params}

    @property
    def _client_params(self) -> Mapping[str, Any]:
        """Get the parameters used for the openai client."""
        zhipu_creds: Dict[str, Any] = {
            'access_key_id': self.access_key_id,
            'secret_access_key': self.secret_access_key,
            'model': self.model_name,
        }
        return {**zhipu_creds, **self._default_params}

    def _get_invocation_params(self,
                               stop: Optional[List[str]] = None,
                               **kwargs: Any) -> Dict[str, Any]:
        """Get the parameters used to invoke the model FOR THE CALLBACKS."""
        return {
            **super()._get_invocation_params(stop=stop, **kwargs),
            **self._default_params,
            'model': self.model_name,
            'function': kwargs.get('functions'),
        }

    @property
    def _llm_type(self) -> str:
        """Return type of chat model."""
        return 'sense-chat'



"""proxy llm chat wrapper."""
from __future__ import annotations

import logging
import sys
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Mapping, Optional, Tuple, Union

# import requests
from langchain.callbacks.manager import AsyncCallbackManagerForLLMRun, CallbackManagerForLLMRun
from langchain.chat_models.base import BaseChatModel
from langchain.schema import ChatGeneration, ChatResult
from langchain.schema.messages import (AIMessage, BaseMessage, ChatMessage, FunctionMessage,
                                       HumanMessage, SystemMessage)
from langchain.utils import get_from_dict_or_env
from langchain_core.pydantic_v1 import Field, root_validator
from tenacity import (before_sleep_log, retry, retry_if_exception_type, stop_after_attempt,
                      wait_exponential)

from .interface import XunfeiChatCompletion
from .interface.types import ChatInput

if TYPE_CHECKING:
    import tiktoken

logger = logging.getLogger(__name__)


def _import_tiktoken() -> Any:
    try:
        import tiktoken
    except ImportError:
        raise ValueError('Could not import tiktoken python package. '
                         'This is needed in order to calculate get_token_ids. '
                         'Please install it with `pip install tiktoken`.')
    return tiktoken


def _create_retry_decorator(llm: ChatXunfeiAI) -> Callable[[Any], Any]:

    min_seconds = 1
    max_seconds = 20
    # Wait 2^x * 1 second between each retry starting with
    # 4 seconds, then up to 10 seconds, then 10 seconds afterwards
    return retry(
        reraise=True,
        stop=stop_after_attempt(llm.max_retries),
        wait=wait_exponential(multiplier=1, min=min_seconds, max=max_seconds),
        retry=(retry_if_exception_type(Exception)),
        before_sleep=before_sleep_log(logger, logging.WARNING),
    )


def _convert_dict_to_message(_dict: Mapping[str, Any]) -> BaseMessage:
    role = _dict['role']
    if role == 'user':
        return HumanMessage(content=_dict['content'])
    elif role == 'assistant':
        content = _dict[
            'content'] or ''  # OpenAI returns None for tool invocations
        if _dict.get('function_call'):
            additional_kwargs = {'function_call': dict(_dict['function_call'])}
        else:
            additional_kwargs = {}
        return AIMessage(content=content, additional_kwargs=additional_kwargs)
    elif role == 'system':
        return SystemMessage(content=_dict['content'])
    elif role == 'function':
        return FunctionMessage(content=_dict['content'], name=_dict['name'])
    else:
        return ChatMessage(content=_dict['content'], role=role)


def _convert_message_to_dict(message: BaseMessage) -> dict:
    if isinstance(message, ChatMessage):
        message_dict = {'role': message.role, 'content': message.content}
    elif isinstance(message, HumanMessage):
        message_dict = {'role': 'user', 'content': message.content}
    elif isinstance(message, AIMessage):
        message_dict = {'role': 'assistant', 'content': message.content}
        if 'function_call' in message.additional_kwargs:
            message_dict['function_call'] = message.additional_kwargs[
                'function_call']
    elif isinstance(message, SystemMessage):
        message_dict = {'role': 'system', 'content': message.content}
    elif isinstance(message, FunctionMessage):
        message_dict = {
            'role': 'function',
            'content': message.content,
            'name': message.name,
        }
    else:
        raise ValueError(f'Got unknown type {message}')
    if 'name' in message.additional_kwargs:
        message_dict['name'] = message.additional_kwargs['name']
    return message_dict


class ChatXunfeiAI(BaseChatModel):
    """Wrapper around proxy Chat large language models.

    To use, the environment variable ``ELEMAI_API_KEY`` set with your API key.

    Example:
        .. code-block:: python

            from bisheng_langchain.chat_models import ChatXunfeiAI
            chat_miniamaxai = ChatXunfeiAI(model_name="spark")
    """

    client: Optional[Any]  #: :meta private:
    """Model name to use."""
    model_name: str = Field('spark-1.5', alias='model')

    temperature: float = 0.5
    top_p: float = 0.7
    """What sampling temperature to use."""
    model_kwargs: Optional[Dict[str, Any]] = Field(default_factory=dict)
    """Holds any model parameters valid for `create` call not explicitly specified."""
    xunfeiai_appid: Optional[str] = None
    xunfeiai_api_key: Optional[str] = None
    xunfeiai_api_secret: Optional[str] = None

    request_timeout: Optional[Union[float, Tuple[float, float]]] = None
    """Timeout for requests to OpenAI completion API. Default is 600 seconds."""
    max_retries: Optional[int] = 6
    """Maximum number of retries to make when generating."""
    streaming: Optional[bool] = False
    """Whether to stream the results or not."""
    n: Optional[int] = 1
    """Number of chat completions to generate for each prompt."""
    max_tokens: Optional[int] = 1024
    """Maximum number of tokens to generate."""
    tiktoken_model_name: Optional[str] = None
    """The model name to pass to tiktoken when using this class.
    Tiktoken is used to count the number of tokens in documents to constrain
    them to be under a certain limit. By default, when set to None, this will
    be the same as the embedding model name. However, there are some cases
    where you may want to use this Embedding class with a model name not
    supported by tiktoken. This can include when using Azure embeddings or
    when using one of the many model providers that expose an OpenAI-like
    API but with different models. In those cases, in order to avoid erroring
    when tiktoken is called, you can specify a model name to use here."""
    verbose: Optional[bool] = False

    class Config:
        """Configuration for this pydantic object."""

        allow_population_by_field_name = True

    @root_validator()
    def validate_environment(cls, values: Dict) -> Dict:
        """Validate that api key and python package exists in environment."""
        values['xunfeiai_appid'] = get_from_dict_or_env(
            values, 'xunfeiai_appid', 'XUNFEIAI_APPID')

        values['xunfeiai_api_key'] = get_from_dict_or_env(
            values, 'xunfeiai_api_key', 'XUNFEIAI_API_KEY')

        values['xunfeiai_api_secret'] = get_from_dict_or_env(
            values, 'xunfeiai_api_secret', 'XUNFEIAI_API_SECRET')

        appid = values['xunfeiai_appid']
        api_key = values['xunfeiai_api_key']
        api_secret = values['xunfeiai_api_secret']

        try:
            values['client'] = XunfeiChatCompletion(appid, api_key, api_secret)
        except AttributeError:
            raise ValueError(
                'Try upgrading it with `pip install --upgrade requests`.')
        return values

    @property
    def _default_params(self) -> Dict[str, Any]:
        """Get the default parameters for calling ChatXunfeiAI API."""
        return {
            'model': self.model_name,
            'temperature': self.temperature,
            'top_p': self.top_p,
            'max_tokens': self.max_tokens,
            **self.model_kwargs,
        }

    def completion_with_retry(self, **kwargs: Any) -> Any:
        retry_decorator = _create_retry_decorator(self)

        @retry_decorator
        def _completion_with_retry(**kwargs: Any) -> Any:
            messages = kwargs.get('messages')
            temperature = kwargs.get('temperature')
            top_p = kwargs.get('top_p')
            max_tokens = kwargs.get('max_tokens')
            params = {
                'messages': messages,
                'model': self.model_name,
                'top_p': top_p,
                'temperature': temperature,
                'max_tokens': max_tokens
            }
            return self.client(ChatInput.parse_obj(params),
                               self.verbose).dict()

        return _completion_with_retry(**kwargs)

    def _combine_llm_outputs(self, llm_outputs: List[Optional[dict]]) -> dict:
        overall_token_usage: dict = {}
        for output in llm_outputs:
            if output is None:
                # Happens in streaming
                continue
            token_usage = output['token_usage']
            for k, v in token_usage.items():
                if k in overall_token_usage:
                    overall_token_usage[k] += v
                else:
                    overall_token_usage[k] = v
        return {
            'token_usage': overall_token_usage,
            'model_name': self.model_name
        }

    def _generate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> ChatResult:
        message_dicts, params = self._create_message_dicts(messages, stop)
        params = {**params, **kwargs}

        response = self.completion_with_retry(messages=message_dicts, **params)
        return self._create_chat_result(response)

    async def _agenerate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> ChatResult:
        return self._generate(messages, stop, run_manager, **kwargs)

    def _create_message_dicts(
        self, messages: List[BaseMessage], stop: Optional[List[str]]
    ) -> Tuple[List[Dict[str, Any]], Dict[str, Any]]:
        params = dict(self._client_params)
        if stop is not None:
            if 'stop' in params:
                raise ValueError(
                    '`stop` found in both the input and default params.')
            params['stop'] = stop

        message_dicts = [_convert_message_to_dict(m) for m in messages]

        return message_dicts, params

    def _create_chat_result(self, response: Mapping[str, Any]) -> ChatResult:
        generations = []
        for res in response['choices']:
            message = _convert_dict_to_message(res['message'])
            gen = ChatGeneration(message=message)
            generations.append(gen)

        llm_output = {
            'token_usage': response['usage'],
            'model_name': self.model_name
        }
        return ChatResult(generations=generations, llm_output=llm_output)

    @property
    def _identifying_params(self) -> Mapping[str, Any]:
        """Get the identifying parameters."""
        return {**{'model_name': self.model_name}, **self._default_params}

    @property
    def _client_params(self) -> Mapping[str, Any]:
        """Get the parameters used for the client."""
        minimaxai_creds: Dict[str, Any] = {
            'model': self.model_name,
        }
        return {**minimaxai_creds, **self._default_params}

    def _get_invocation_params(self,
                               stop: Optional[List[str]] = None,
                               **kwargs: Any) -> Dict[str, Any]:
        """Get the parameters used to invoke the model FOR THE CALLBACKS."""
        return {
            **super()._get_invocation_params(stop=stop, **kwargs),
            **self._default_params,
            'model': self.model_name,
            'function': kwargs.get('functions'),
        }

    @property
    def _llm_type(self) -> str:
        """Return type of chat model."""
        return 'xunfeiai_chat'

    def _get_encoding_model(self) -> Tuple[str, tiktoken.Encoding]:
        tiktoken_ = _import_tiktoken()
        if self.tiktoken_model_name is not None:
            model = self.tiktoken_model_name
        else:
            model = self.model_name
            # model chatglm-std, chatglm-lite
        # Returns the number of tokens used by a list of messages.
        try:
            encoding = tiktoken_.encoding_for_model(model)
        except KeyError:
            logger.warning(
                'Warning: model not found. Using cl100k_base encoding.')
            model = 'cl100k_base'
            encoding = tiktoken_.get_encoding(model)
        return model, encoding

    def get_token_ids(self, text: str) -> List[int]:
        """Get the tokens present in the text with tiktoken package."""
        # tiktoken NOT supported for Python 3.7 or below
        if sys.version_info[1] <= 7:
            return super().get_token_ids(text)
        _, encoding_model = self._get_encoding_model()
        return encoding_model.encode(text)

    def get_num_tokens_from_messages(self, messages: List[BaseMessage]) -> int:
        """Calculate num tokens for chatglm with tiktoken package.

        todo: read chatglm document
        Official documentation: https://github.com/openai/openai-cookbook/blob/
        main/examples/How_to_format_inputs_to_ChatGPT_models.ipynb"""
        if sys.version_info[1] <= 7:
            return super().get_num_tokens_from_messages(messages)
        model, encoding = self._get_encoding_model()
        if model.startswith('spark'):
            # every message follows <im_start>{role/name}\n{content}<im_end>\n
            tokens_per_message = 4
            # if there's a name, the role is omitted
            tokens_per_name = -1
        else:
            raise NotImplementedError(
                f'get_num_tokens_from_messages() is not presently implemented '
                f'for model {model}.'
                'See https://github.com/openai/openai-python/blob/main/chatml.md for '
                'information on how messages are converted to tokens.')
        num_tokens = 0
        messages_dict = [_convert_message_to_dict(m) for m in messages]
        for message in messages_dict:
            num_tokens += tokens_per_message
            for key, value in message.items():
                num_tokens += len(encoding.encode(value))
                if key == 'name':
                    num_tokens += tokens_per_name
        # every reply is primed with <im_start>assistant
        num_tokens += 3
        return num_tokens



import argparse
import copy
import inspect
import time
import os
from collections import defaultdict

import httpx
import pandas as pd
import yaml
from loguru import logger
from tqdm import tqdm
from bisheng_langchain.retrievers import EnsembleRetriever
from bisheng_langchain.vectorstores import ElasticKeywordsSearch, Milvus
from langchain.chains.question_answering import load_qa_chain
from bisheng_langchain.rag.init_retrievers import (
    BaselineVectorRetriever,
    KeywordRetriever,
    MixRetriever,
    SmallerChunksVectorRetriever,
)
from bisheng_langchain.rag.scoring.ragas_score import RagScore
from bisheng_langchain.rag.utils import import_by_type, import_class


class BishengRagPipeline:

    def __init__(self, yaml_path) -> None:
        self.yaml_path = yaml_path
        with open(self.yaml_path, 'r') as f:
            self.params = yaml.safe_load(f)

        # init data
        self.origin_file_path = self.params['data']['origin_file_path']
        self.question_path = self.params['data']['question']
        self.save_answer_path = self.params['data']['save_answer']

        # init embeddings
        embedding_params = self.params['embedding']
        embedding_object = import_by_type(_type='embeddings', name=embedding_params['type'])
        if embedding_params['type'] == 'OpenAIEmbeddings' and embedding_params['openai_proxy']:
            embedding_params.pop('type')
            self.embeddings = embedding_object(
                http_client=httpx.Client(proxies=embedding_params['openai_proxy']), **embedding_params
            )
        else:
            embedding_params.pop('type')
            self.embeddings = embedding_object(**embedding_params)

        # init llm
        llm_params = self.params['chat_llm']
        llm_object = import_by_type(_type='llms', name=llm_params['type'])
        if llm_params['type'] == 'ChatOpenAI' and llm_params['openai_proxy']:
            llm_params.pop('type')
            self.llm = llm_object(http_client=httpx.Client(proxies=llm_params['openai_proxy']), **llm_params)
        else:
            llm_params.pop('type')
            self.llm = llm_object(**llm_params)

        # milvus
        self.vector_store = Milvus(
            embedding_function=self.embeddings,
            connection_args={
                "host": self.params['milvus']['host'],
                "port": self.params['milvus']['port'],
            },
        )

        # es
        self.keyword_store = ElasticKeywordsSearch(
            index_name='default_es',
            elasticsearch_url=self.params['elasticsearch']['url'],
            ssl_verify=self.params['elasticsearch']['ssl_verify'],
        )

        # init retriever
        retriever_list = []
        retrievers = self.params['retriever']['retrievers']
        for retriever in retrievers:
            retriever_type = retriever.pop('type')
            retriever_params = {
                'vector_store': self.vector_store,
                'keyword_store': self.keyword_store,
                'splitter_kwargs': retriever['splitter'],
                'retrieval_kwargs': retriever['retrieval'],
            }
            retriever_list.append(self._post_init_retriever(retriever_type=retriever_type, **retriever_params))
        self.retriever = EnsembleRetriever(retrievers=retriever_list)

    def _post_init_retriever(self, retriever_type, **kwargs):
        retriever_classes = {
            'KeywordRetriever': KeywordRetriever,
            'BaselineVectorRetriever': BaselineVectorRetriever,
            'MixRetriever': MixRetriever,
            'SmallerChunksVectorRetriever': SmallerChunksVectorRetriever,
        }
        if retriever_type not in retriever_classes:
            raise ValueError(f'Unknown retriever type: {retriever_type}')

        input_kwargs = {}
        splitter_params = kwargs.pop('splitter_kwargs')
        for key, value in splitter_params.items():
            splitter_obj = import_by_type(_type='textsplitters', name=value.pop('type'))
            input_kwargs[key] = splitter_obj(**value)

        retrieval_params = kwargs.pop('retrieval_kwargs')
        for key, value in retrieval_params.items():
            input_kwargs[key] = value

        input_kwargs['vector_store'] = kwargs.pop('vector_store')
        input_kwargs['keyword_store'] = kwargs.pop('keyword_store')

        retriever_class = retriever_classes[retriever_type]
        return retriever_class(**input_kwargs)

    def file2knowledge(self):
        """
        file to knowledge
        """
        df = pd.read_excel(self.question_path)
        if ('文件名' not in df.columns) or ('知识库名' not in df.columns):
            raise Exception(f'文件名 or 知识库名 not in {self.question_path}.')

        loader_params = self.params['loader']
        loader_object = import_by_type(_type='documentloaders', name=loader_params.pop('type'))

        all_questions_info = df.to_dict('records')
        collectionname2filename = defaultdict(set)
        for info in all_questions_info:
            # 存入set，去掉重复的文件名
            collectionname2filename[info['知识库名']].add(info['文件名'])

        for collection_name in tqdm(collectionname2filename):
            all_file_paths = []
            for file_name in collectionname2filename[collection_name]:
                file_path = os.path.join(self.origin_file_path, file_name)
                if not os.path.exists(file_path):
                    raise Exception(f'{file_path} not exists.')
                # file path可以是文件夹或者单个文件
                if os.path.isdir(file_path):
                    # 文件夹包含多个文件
                    all_file_paths.extend(
                        [os.path.join(file_path, name) for name in os.listdir(file_path) if not name.startswith('.')]
                    )
                else:
                    # 单个文件
                    all_file_paths.append(file_path)

            # 当前知识库需要存储的所有文件
            collection_name = f"{collection_name}_{self.params['retriever']['suffix']}"
            for index, each_file_path in enumerate(all_file_paths):
                logger.info(f'each_file_path: {each_file_path}')
                loader = loader_object(
                    file_name=os.path.basename(each_file_path), file_path=each_file_path, **loader_params
                )
                documents = loader.load()
                logger.info(f'documents: {len(documents)}')
                if len(documents[0].page_content) == 0:
                    logger.error(f'{each_file_path} page_content is empty.')

                vector_drop_old = self.params['milvus']['drop_old'] if index == 0 else False
                keyword_drop_old = self.params['elasticsearch']['drop_old'] if index == 0 else False
                for idx, retriever in enumerate(self.retriever.retrievers):
                    retriever.add_documents(documents, f"{collection_name}_{idx}", vector_drop_old)
                    # retriever.add_documents(documents, collection_name, vector_drop_old)

    def retrieval_and_rerank(self, question, collection_name):
        """
        retrieval and rerank
        """
        collection_name = f"{collection_name}_{self.params['retriever']['suffix']}"

        # EnsembleRetriever直接检索召回会默认去重
        # docs = self.retriever.get_relevant_documents(query=question, collection_name=collection_name)
        docs = []
        for idx, retriever in enumerate(self.retriever.retrievers):
            docs.extend(retriever.get_relevant_documents(query=question, collection_name=f"{collection_name}_{idx}"))
            # docs.extend(retriever.get_relevant_documents(query=question, collection_name=collection_name))
        logger.info(f'retrieval docs: {len(docs)}')

        # delete duplicate
        if self.params['post_retrieval']['delete_duplicate']:
            logger.info(f'origin docs: {len(docs)}')
            all_contents = []
            docs_no_dup = []
            for index, doc in enumerate(docs):
                doc_content = doc.page_content
                if doc_content in all_contents:
                    continue
                all_contents.append(doc_content)
                docs_no_dup.append(doc)
            docs = docs_no_dup
            logger.info(f'delete duplicate docs: {len(docs)}')

        # rerank
        if self.params['post_retrieval']['with_rank'] and len(docs):
            if not hasattr(self, 'ranker'):
                rerank_params = self.params['post_retrieval']['rerank']
                rerank_type = rerank_params.pop('type')
                rerank_object = import_class(f'bisheng_langchain.rag.rerank.{rerank_type}')
                self.ranker = rerank_object(**rerank_params)
            docs = getattr(self, 'ranker').sort_and_filter(question, docs)

        return docs

    def load_documents(self, file_name, max_content=100000):
        """
        max_content: max content len of llm
        """
        file_path = os.path.join(self.origin_file_path, file_name)
        if not os.path.exists(file_path):
            raise Exception(f'{file_path} not exists.')
        if os.path.isdir(file_path):
            raise Exception(f'{file_path} is a directory.')

        loader_params = copy.deepcopy(self.params['loader'])
        loader_object = import_by_type(_type='documentloaders', name=loader_params.pop('type'))
        loader = loader_object(file_name=file_name, file_path=file_path, **loader_params)

        documents = loader.load()
        logger.info(f'documents: {len(documents)}, page_content: {len(documents[0].page_content)}')
        for doc in documents:
            doc.page_content = doc.page_content[:max_content]
        return documents

    def question_answering(self):
        """
        question answer over knowledge
        """
        df = pd.read_excel(self.question_path)
        all_questions_info = df.to_dict('records')
        if 'prompt_type' in self.params['generate']:
            prompt_type = self.params['generate']['prompt_type']
            prompt = import_class(f'bisheng_langchain.rag.prompts.{prompt_type}')
        else:
            prompt = None
        qa_chain = load_qa_chain(
            llm=self.llm, chain_type=self.params['generate']['chain_type'], prompt=prompt, verbose=False
        )
        file2docs = dict()
        for questions_info in tqdm(all_questions_info):
            question = questions_info['问题']
            file_name = questions_info['文件名']
            collection_name = questions_info['知识库名']

            if self.params['generate']['with_retrieval']:
                # retrieval and rerank
                docs = self.retrieval_and_rerank(question, collection_name)
            else:
                # load document
                if file_name not in file2docs:
                    docs = self.load_documents(file_name)
                    file2docs[file_name] = docs
                else:
                    docs = file2docs[file_name]

            # question answer
            try:
                ans = qa_chain({"input_documents": docs, "question": question}, return_only_outputs=False)
            except Exception as e:
                logger.error(f'question: {question}\nerror: {e}')
                ans = {'output_text': str(e)}

            # context = '\n\n'.join([doc.page_content for doc in docs])
            # content = prompt.format(context=context, question=question)

            # # for rate_limit
            # time.sleep(15)

            rag_answer = ans['output_text']
            logger.info(f'question: {question}\nans: {rag_answer}\n')
            questions_info['rag_answer'] = rag_answer
            # questions_info['rag_context'] = '\n----------------\n'.join([doc.page_content for doc in docs])
            # questions_info['rag_context'] = content

        df = pd.DataFrame(all_questions_info)
        df.to_excel(self.save_answer_path, index=False)

    def score(self):
        """
        score
        """
        metric_params = self.params['metric']
        if metric_params['type'] == 'bisheng-ragas':
            score_params = {
                'excel_path': self.save_answer_path,
                'save_path': os.path.dirname(self.save_answer_path),
                'question_column': metric_params['question_column'],
                'gt_column': metric_params['gt_column'],
                'answer_column': metric_params['answer_column'],
                'query_type_column': metric_params.get('query_type_column', None),
                'contexts_column': metric_params.get('contexts_column', None),
                'metrics': metric_params['metrics'],
                'batch_size': metric_params['batch_size'],
                'gt_split_column': metric_params.get('gt_split_column', None),
                'whether_gtsplit': metric_params.get('whether_gtsplit', False), # 是否需要模型对gt进行要点拆分
            }
            rag_score = RagScore(**score_params)
            rag_score.score()
        else:
            # todo: 其他评分方法
            pass


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Process some integers.')
    # 添加参数
    parser.add_argument('--mode', type=str, default='qa', help='upload or qa or score')
    parser.add_argument('--params', type=str, default='config/test/baseline_s2b.yaml', help='bisheng rag params')
    # 解析参数
    args = parser.parse_args()

    rag = BishengRagPipeline(args.params)

    if args.mode == 'upload':
        rag.file2knowledge()
    elif args.mode == 'qa':
        rag.question_answering()
    elif args.mode == 'score':
        rag.score()



from typing import List

from langchain.callbacks.manager import (AsyncCallbackManagerForRetrieverRun,
                                         CallbackManagerForRetrieverRun)
from langchain.schema import BaseRetriever, Document


class MixEsVectorRetriever(BaseRetriever):
    """
    This class ensemble the results of es retriever and vector retriever.

    Args:
        retrievers: A list of retrievers to ensemble.
        weights: A list of weights corresponding to the retrievers. Defaults to equal
            weighting for all retrievers.
        c: A constant added to the rank, controlling the balance between the importance
            of high-ranked items and the consideration given to lower-ranked items.
            Default is 60.
    """

    vector_retriever: BaseRetriever
    keyword_retriever: BaseRetriever
    combine_strategy: str = 'keyword_front'  # "keyword_front, vector_front, mix"

    def _get_relevant_documents(
        self,
        query: str,
        *,
        run_manager: CallbackManagerForRetrieverRun,
    ) -> List[Document]:
        """
        Get the relevant documents for a given query.

        Args:
            query: The query to search for.

        Returns:
            A list of documents.
        """

        # Get fused result of the retrievers.
        vector_docs = self.vector_retriever.get_relevant_documents(
            query, callbacks=run_manager.get_child())
        keyword_docs = self.keyword_retriever.get_relevant_documents(
            query, callbacks=run_manager.get_child())

        if self.combine_strategy == 'keyword_front':
            return keyword_docs + vector_docs
        elif self.combine_strategy == 'vector_front':
            return vector_docs + keyword_docs
        elif self.combine_strategy == 'mix':
            combine_docs_dict = {}
            min_len = min(len(keyword_docs), len(vector_docs))
            for i in range(min_len):
                combine_docs_dict[keyword_docs[i].page_content] = keyword_docs[i]
                combine_docs_dict[vector_docs[i].page_content] = vector_docs[i]
            for doc in keyword_docs[min_len:]:
                combine_docs_dict[doc.page_content] = doc
            for doc in vector_docs[min_len:]:
                combine_docs_dict[doc.page_content] = doc

            # 将字典的值转换为列表
            combine_docs = list(combine_docs_dict.values())
            return combine_docs
        else:
            raise ValueError(f'Expected combine_strategy to be one of '
                             f'(keyword_front, vector_front, mix),'
                             f'instead found {self.combine_strategy}')

    async def _aget_relevant_documents(
        self,
        query: str,
        *,
        run_manager: AsyncCallbackManagerForRetrieverRun,
    ) -> List[Document]:
        """
        Asynchronously get the relevant documents for a given query.

        Args:
            query: The query to search for.

        Returns:
            A list of reranked documents.
        """

        # Get fused result of the retrievers.
        vector_docs = await self.vector_retriever.aget_relevant_documents(
            query, callbacks=run_manager.get_child())
        keyword_docs = await self.keyword_retriever.aget_relevant_documents(
            query, callbacks=run_manager.get_child())
        if self.combine_strategy == 'keyword_front':
            return keyword_docs + vector_docs
        elif self.combine_strategy == 'vector_front':
            return vector_docs + keyword_docs
        elif self.combine_strategy == 'mix':
            combine_docs_dict = {}
            min_len = min(len(keyword_docs), len(vector_docs))
            for i in range(min_len):
                combine_docs_dict[keyword_docs[i].page_content] = keyword_docs[i]
                combine_docs_dict[vector_docs[i].page_content] = vector_docs[i]
            for doc in keyword_docs[min_len:]:
                combine_docs_dict[doc.page_content] = doc
            for doc in vector_docs[min_len:]:
                combine_docs_dict[doc.page_content] = doc

            # 将字典的值转换为列表
            combine_docs = list(combine_docs_dict.values())
            return combine_docs
        else:
            raise ValueError(f'Expected combine_strategy to be one of '
                             f'(keyword_front, vector_front, mix),'
                             f'instead found {self.combine_strategy}')



from langchain_core.prompts import PromptTemplate
from langchain_core.prompts.chat import (
    ChatPromptTemplate,
    HumanMessagePromptTemplate,
    SystemMessagePromptTemplate,
)


prompt_template = """Use the following pieces of context to answer the question at the end. If you don't know the answer, just say that you don't know, don't try to make up an answer.

{context}

Question: {question}
Helpful Answer:"""
BASE_PROMPT = PromptTemplate(
    template=prompt_template, input_variables=["context", "question"]
)


system_template = """Use the following pieces of context to answer the user's question. 
If you don't know the answer, just say that you don't know, don't try to make up an answer.
----------------
{context}"""
messages = [
    SystemMessagePromptTemplate.from_template(system_template),
    HumanMessagePromptTemplate.from_template("{question}"),
]
CHAT_PROMPT = ChatPromptTemplate.from_messages(messages)


system_template_general = """你是一个准确且可靠的知识库问答助手，能够借助上下文知识回答问题。你需要根据以下的规则来回答问题：
1. 如果上下文中包含了正确答案，你需要根据上下文进行准确的回答。但是在回答前，你需要注意，上下文中的信息可能存在事实性错误，如果文档中存在和事实不一致的错误，请根据事实回答。
2. 如果上下文中不包含答案，就说你不知道，不要试图编造答案。
3. 你需要根据上下文给出详细的回答，不要试图偷懒，不要遗漏括号中的信息，你必须回答的尽可能详细。
"""
human_template_general = """
上下文：
{context}

问题：
{question}
"""
messages_general = [
    SystemMessagePromptTemplate.from_template(system_template_general),
    HumanMessagePromptTemplate.from_template(human_template_general),
]
CHAT_PROMPT_GENERAL = ChatPromptTemplate.from_messages(messages_general)


import os
import uuid
from loguru import logger
from dataclasses import dataclass
from typing import Any, Dict, Iterable, List, Optional

from bisheng_langchain.vectorstores import ElasticKeywordsSearch
from bisheng_langchain.vectorstores.milvus import Milvus
from langchain_core.documents import Document
from langchain_core.pydantic_v1 import Field
from langchain_core.retrievers import BaseRetriever
from langchain_core.vectorstores import VectorStore

from langchain.callbacks.manager import CallbackManagerForRetrieverRun
from langchain.text_splitter import TextSplitter


class KeywordRetriever(BaseRetriever):
    keyword_store: VectorStore
    text_splitter: TextSplitter
    search_type: str = 'similarity'
    search_kwargs: dict = Field(default_factory=dict)

    def add_documents(
        self,
        documents: List[Document],
        collection_name: str,
        drop_old: bool = False,
        **kwargs,
    ) -> None:
        split_docs = self.text_splitter.split_documents(documents)
        logger.info(f"KeywordRetriever: split document into {len(split_docs)} chunks")
        for chunk_index, split_doc in enumerate(split_docs):
            if 'chunk_bboxes' in split_doc.metadata:
                split_doc.metadata.pop('chunk_bboxes')
            split_doc.metadata['chunk_index'] = chunk_index
            if kwargs.get('add_aux_info', False):
                split_doc.page_content = split_doc.metadata["source"] + '\n' + split_doc.metadata["title"] + '\n' + split_doc.page_content

        elasticsearch_url = self.keyword_store.elasticsearch_url
        ssl_verify = self.keyword_store.ssl_verify
        self.keyword_store.from_documents(
            split_docs,
            embedding='',
            index_name=collection_name,
            elasticsearch_url=elasticsearch_url,
            ssl_verify=ssl_verify,
            drop_old=drop_old,
        )

    def _get_relevant_documents(
        self,
        query: str,
        collection_name: Optional[str] = None,
    ) -> List[Document]:
        if collection_name:
            self.keyword_store = self.keyword_store.__class__(
                index_name=collection_name,
                elasticsearch_url=self.keyword_store.elasticsearch_url,
                ssl_verify=self.keyword_store.ssl_verify,
                llm_chain=self.keyword_store.llm_chain
            )
        if self.search_type == 'similarity':
            result = self.keyword_store.similarity_search(query, **self.search_kwargs)
        return result



"""Chain that runs an arbitrary python function."""
import functools
import json
import logging
from typing import Any, Dict, List, Optional

from autogen import ConversableAgent
from bisheng_langchain.autogen_role import AutoGenGroupChatManager, AutoGenUser
from langchain.callbacks.manager import AsyncCallbackManagerForChainRun, CallbackManagerForChainRun
from langchain.chains.base import Chain

logger = logging.getLogger(__name__)


class AutoGenChain(Chain):
    """Chain that print the loader output.
    """
    user_proxy_agent: AutoGenUser
    recipient: ConversableAgent

    input_key: str = 'chat_topic'  #: :meta private:

    output_key: str = 'chat_content'  #: :meta private:

    @staticmethod
    @functools.lru_cache
    def _log_once(msg: str) -> None:
        """Log a message once.

        :meta private:
        """
        logger.warning(msg)

    @property
    def input_keys(self) -> List[str]:
        """Expect input keys.

        :meta private:
        """
        return [self.input_key]

    @property
    def output_keys(self) -> List[str]:
        """Return output keys.

        :meta private:
        """
        return [self.output_key]

    def _call(
        self,
        inputs: Dict[str, str],
        run_manager: Optional[CallbackManagerForChainRun] = None,
    ) -> Dict[str, str]:
        message = inputs[self.input_key]
        # io_output = io.StringIO()
        # with contextlib.redirect_stdout(io_output):
        global_chat_messages = []
        self.user_proxy_agent.initiate_chat(self.recipient, message=message,
                                            global_chat_messages=global_chat_messages,
                                            run_manager=run_manager)
        # chat_content = io_output.getvalue()
        chat_content = json.dumps(
            global_chat_messages, indent=2, ensure_ascii=False)
        output = {self.output_key: chat_content}
        return output

    async def _acall(
        self,
        inputs: Dict[str, Any],
        run_manager: Optional[AsyncCallbackManagerForChainRun] = None,
    ) -> Dict[str, Any]:
        message = inputs[self.input_key]
        # io_output = io.StringIO()
        # with contextlib.redirect_stdout(io_output):
        global_chat_messages = []
        self.recipient.stop = False
        await self.user_proxy_agent.a_initiate_chat(self.recipient,
                                                    message=message,
                                                    global_chat_messages=global_chat_messages,
                                                    run_manager=run_manager)
        # chat_content = io_output.getvalue()
        output = {self.output_key: global_chat_messages[-1].get('message'),
                  'intermediate_steps': global_chat_messages}
        return output

    async def stop(self):
        self.recipient.stop = True
        self.user_proxy_agent.event.set()
        self.user_proxy_agent.event.clear()

    async def reset(self):
        if isinstance(self.recipient, AutoGenGroupChatManager):
            self.recipient.reset()

    async def input(self, input: str):
        self.user_proxy_agent.input = input
        self.user_proxy_agent.event.set()
        self.user_proxy_agent.event.clear()



from .elastic_keywords_search import ElasticKeywordsSearch
from .milvus import Milvus
from .retriever import VectorStoreFilterRetriever
from .JiShiJiaoMilvus import JiShiJiaoMilvus
from .JiShiJiao_elastic_keywords_search import JiShiJiaoElasticKeywordsSearch

__all__ = ['ElasticKeywordsSearch', 'VectorStoreFilterRetriever', 'Milvus', "JiShiJiaoMilvus",
           "JiShiJiaoElasticKeywordsSearch"]



from typing import Any, Dict, Iterable, List, Optional

from bisheng_langchain.vectorstores import ElasticKeywordsSearch
from bisheng_langchain.vectorstores.milvus import Milvus
from langchain_core.documents import Document
from langchain_core.pydantic_v1 import Field
from langchain_core.retrievers import BaseRetriever

from langchain.schema import BaseRetriever, Document
from langchain.text_splitter import TextSplitter


class MixRetriever(BaseRetriever):
    vector_store: Milvus
    keyword_store: ElasticKeywordsSearch
    vector_text_splitter: TextSplitter
    keyword_text_splitter: TextSplitter
    # retrieval
    search_type: str = 'similarity'
    vector_search_kwargs: dict = Field(default_factory=dict)
    keyword_search_kwargs: dict = Field(default_factory=dict)
    combine_strategy: str = 'mix'  # "keyword_front, vector_front, mix"

    def add_documents(
        self,
        documents: List[Document],
        collection_name: str,
        drop_old: bool = False,
        **kwargs,
    ) -> None:
        vector_split_docs = self.vector_text_splitter.split_documents(documents)
        for chunk_index, split_doc in enumerate(vector_split_docs):
            if 'chunk_bboxes' in split_doc.metadata:
                split_doc.metadata.pop('chunk_bboxes')
            split_doc.metadata['chunk_index'] = chunk_index
            if kwargs.get('add_aux_info', False):
                split_doc.page_content = split_doc.metadata["source"] + '\n' + split_doc.metadata["title"] + '\n' + split_doc.page_content
        keyword_split_docs = self.keyword_text_splitter.split_documents(documents)
        for chunk_index, split_doc in enumerate(keyword_split_docs):
            if 'chunk_bboxes' in split_doc.metadata:
                split_doc.metadata.pop('chunk_bboxes')
            split_doc.metadata['chunk_index'] = chunk_index
            if kwargs.get('add_aux_info', False):
                split_doc.page_content = split_doc.metadata["source"] + '\n' + split_doc.metadata["title"] + '\n' + split_doc.page_content

        self.keyword_store.from_documents(
            keyword_split_docs,
            embedding='',
            index_name=collection_name,
            elasticsearch_url=self.keyword_store.elasticsearch_url,
            ssl_verify=self.keyword_store.ssl_verify,
            drop_old=drop_old,
        )

        self.vector_store.from_documents(
            vector_split_docs,
            embedding=self.vector_store.embedding_func,
            collection_name=collection_name,
            connection_args=self.vector_store.connection_args,
            drop_old=drop_old,
        )

    def _get_relevant_documents(
        self,
        query: str,
        collection_name: Optional[str] = None,
    ) -> List[Document]:
        if collection_name:
            self.keyword_store = self.keyword_store.__class__(
                index_name=collection_name,
                elasticsearch_url=self.keyword_store.elasticsearch_url,
                ssl_verify=self.keyword_store.ssl_verify,
                llm_chain=self.keyword_store.llm_chain
            )
            self.vector_store = self.vector_store.__class__(
                collection_name=collection_name,
                embedding_function=self.vector_store.embedding_func,
                connection_args=self.vector_store.connection_args,
            )
        if self.search_type == 'similarity':
            keyword_docs = self.keyword_store.similarity_search(query, **self.keyword_search_kwargs)
            vector_docs = self.vector_store.similarity_search(query, **self.vector_search_kwargs)
            if self.combine_strategy == 'keyword_front':
                return keyword_docs + vector_docs
            elif self.combine_strategy == 'vector_front':
                return vector_docs + keyword_docs
            elif self.combine_strategy == 'mix':
                combine_docs = []
                min_len = min(len(keyword_docs), len(vector_docs))
                for i in range(min_len):
                    combine_docs.append(keyword_docs[i])
                    combine_docs.append(vector_docs[i])
                combine_docs.extend(keyword_docs[min_len:])
                combine_docs.extend(vector_docs[min_len:])
                return combine_docs
            else:
                raise ValueError(
                    f'Expected combine_strategy to be one of '
                    f'(keyword_front, vector_front, mix),'
                    f'instead found {self.combine_strategy}'
                )
        else:
            raise ValueError(f'Expected search_type to be one of (similarity), instead found {self.search_type}')



# flake8: noqa  # 禁用 flake8 代码检查
from __future__ import annotations  # 启用类型注解的延迟评估
import bisect  # 导入 bisect 模块，用于二分查找
import copy  # 导入 copy 模块，用于对象复制
import logging  # 导入 logging 模块，用于日志记录
import re  # 导入 re 模块，用于正则表达式操作
from abc import ABC, abstractmethod  # 导入 ABC 和 abstractmethod，用于创建抽象基类和抽象方法
from collections import Counter  # 导入 Counter 类，用于计数
from dataclasses import dataclass  # 导入 dataclass 装饰器，用于创建数据类
from enum import Enum  # 导入 Enum 类，用于创建枚举类型
from typing import (AbstractSet, Any, Callable, Collection, Dict, Iterable, List, Literal, Optional, Sequence, Tuple, Type, TypedDict, TypeVar, Union, cast)  # 导入 typing 模块的各种类型注解
from langchain.docstore.document import Document  # 导入 Langchain 的 Document 类，用于表示文档
from langchain.schema import BaseDocumentTransformer  # 导入 Langchain 的 BaseDocumentTransformer 类，用于文档转换
from langchain.text_splitter import RecursiveCharacterTextSplitter  # 导入 Langchain 的 RecursiveCharacterTextSplitter 类，用于递归字符文本分割
import uuid  # 导入 uuid 模块，用于生成 UUID
logger = logging.getLogger(__name__)  # 获取名为 __name__ 的 logger 实例

def _split_text_with_regex(text: str, separator: str, keep_separator: bool, separator_rule: str) -> List[str]:
    """使用正则表达式分割文本.
    根据提供的分隔符和规则，使用正则表达式分割文本。
    Args:
        text (str): 要分割的文本。
        separator (str): 分隔符正则表达式。
        keep_separator (bool): 是否在分割后的文本块中保留分隔符。
        separator_rule (str): 分隔符规则，可以是 "before" (分隔符在文本块前) 或 "after" (分隔符在文本块后)。
    Returns:
        List[str]: 分割后的文本块列表。
    """
    # 现在我们有了分隔符，分割文本
    if separator:  # 如果分隔符不为空
        if keep_separator:  # 如果需要保留分隔符
            # 模式中的括号使分隔符保留在结果中
            _splits = re.split(f'({separator})', text)  # 使用正则表达式分割文本，分隔符保留在结果列表中
            if separator_rule == "before":  # 如果分隔符规则为 "before"
                splits = [_splits[i] + _splits[i + 1] for i in range(1, len(_splits), 2)]  # 将分隔符添加到前一个文本块
                if len(_splits) % 2 == 0:  # 如果分割后的列表长度为偶数
                    splits += _splits[-1:]  # 添加最后一个文本块
                splits = [_splits[0]] + splits  # 将第一个文本块添加到列表开头
            else:  # 如果分隔符规则为 "after" (默认)
                splits = [_splits[i - 1] + _splits[i] for i in range(1, len(_splits), 2)]  # 将分隔符添加到后一个文本块
                splits = splits + [_splits[-1]]  # 添加最后一个文本块
        else:  # 如果不需要保留分隔符
            splits = re.split(separator, text)  # 使用正则表达式分割文本，分隔符不保留在结果列表中
    else:  # 如果分隔符为空
        splits = list(text)  # 将文本按字符分割成列表
    return [s for s in splits if s != '']  # 移除空字符串

class IntervalSearch(object):
    """间隔搜索类.
    用于在间隔列表中进行搜索，找到给定间隔的索引范围。
    """

    def __init__(self, inters):
        """初始化 IntervalSearch 对象.
        Args:
            inters: 间隔列表，每个间隔表示为一个包含两个元素的列表或元组，例如 [[1, 3], [5, 7], [9, 12]]。
        """
        arrs = []
        for inter in inters:  # 遍历间隔列表
            arrs.extend(inter)  # 将每个间隔的起始和结束点添加到 arrs 列表中
        self.arrs = arrs  # 存储所有间隔的起始和结束点
        self.n = len(self.arrs)  # 存储点的数量

    def _norm_bound(self, ind, v):
        """规范化边界索引.
        根据索引和值，规范化边界索引，确保索引在有效范围内。
        Args:
            ind (int): 边界索引。
            v: 值。
        Returns:
            int: 规范化后的边界索引。
        """
        # [1,3,5,7,9,12]
        # ind=4,8 is empty interval
        # ind=15 is exceed interval
        new_ind = None
        if ind >= self.n:  # 如果索引超出范围
            new_ind = self.n - 1  # 设置为最后一个索引
        elif ind <= 0:  # 如果索引小于等于 0
            new_ind = 0  # 设置为 0
        elif self.arrs[ind] == v:  # 如果索引对应的值等于给定值
            new_ind = ind  # 索引不变
        elif ind % 2 == 0:  # 如果索引为偶数
            if v > self.arrs[ind - 1] and v < self.arrs[ind]:  # 如果值在索引前一个点和当前点之间
                new_ind = ind - 1  # 设置为前一个索引
            else:
                new_ind = ind  # 索引不变
        else:  # 如果索引为奇数
            new_ind = ind  # 索引不变
        return new_ind  # 返回规范化后的索引

    def find(self, inter) -> List[int, int]:
        """查找间隔对应的索引范围.
        在间隔列表中查找给定间隔对应的索引范围。
        Args:
            inter: 给定间隔，例如 [2, 6]。
        Returns:
            List[int, int]: 索引范围列表，包含起始索引和结束索引。
        """
        low_bound1 = bisect.bisect_left(self.arrs, inter[0])  # 二分查找左边界
        low_bound2 = bisect.bisect_left(self.arrs, inter[1])  # 二分查找右边界
        lb1 = self._norm_bound(low_bound1, inter[0])  # 规范化左边界索引
        lb2 = self._norm_bound(low_bound2, inter[1])  # 规范化右边界索引
        return [lb1 // 2, lb2 // 2]  # 返回索引范围，除以 2 是因为 self.arrs 列表是原始间隔列表长度的两倍

class ElemCharacterTextSplitter(RecursiveCharacterTextSplitter):
    """
    ElemCharacterTextSplitter 类，继承自 RecursiveCharacterTextSplitter，用于按字符递归分割文本。
    todo: 类描述待补充
    """

    def __init__(self, separators: Optional[List[str]] = None, separator_rule: Optional[List[str]] = None, is_separator_regex: bool = False, keep_separator: bool = True,
            **kwargs: Any, ) -> None:
        """创建 ElemCharacterTextSplitter 对象.
        Args:
            separators (Optional[List[str]]): 分隔符列表，用于文本分割，默认为 ['\n\n', '\n', ' ', '']。
            separator_rule (Optional[List[str]]): 分隔符规则列表，与 separators 列表对应，定义分隔符的使用方式，默认为 ['after', 'after', 'after', 'after']。
            is_separator_regex (bool): 是否将分隔符视为正则表达式，默认为 False。
            keep_separator (bool): 分割后的文本块是否保留分隔符，默认为 True。
            **kwargs (Any): 传递给 RecursiveCharacterTextSplitter 父类的其他参数。
        """
        super().__init__(separators=separators, keep_separator=keep_separator, **kwargs)
        self._separators = separators or ['\n\n', '\n', ' ', '']  # 设置分隔符列表，默认为 ['\n\n', '\n', ' ', '']
        self._separator_rule = separator_rule or ['after' for _ in range(4)]  # 设置分隔符规则列表，默认为 ['after', 'after', 'after', 'after']
        self.separator_rule = {one: self._separator_rule[index] for index, one in enumerate(separators)}  # 创建分隔符到规则的字典映射
        self._is_separator_regex = is_separator_regex  # 设置是否将分隔符视为正则表达式

    def split_documents(self, documents: Iterable[Document]) -> List[Document]:
        """分割文档列表.
        将文档列表中的每个文档分割成多个文档块。
        Args:
            documents (Iterable[Document]): 包含 Document 对象的 Iterable 对象。
        Returns:
            List[Document]: 分割后的 Document 对象列表。
        """
        texts, metadatas = [], []
        for doc in documents:  # 遍历文档列表
            texts.append(doc.page_content)  # 提取文档内容
            metadatas.append(doc.metadata)  # 提取文档元数据
        return self.create_documents(texts, metadatas=metadatas)  # 调用 create_documents 方法创建分割后的文档

    def _split_text(self, text: str, separators: List[str]) -> List[str]:
        """按分隔符分割文本并返回文本块列表.
        递归地按分隔符分割文本，并返回分割后的文本块列表。
        Args:
            text (str): 要分割的文本。
            separators (List[str]): 分隔符列表。
        Returns:
            List[str]: 分割后的文本块列表。
        """
        final_chunks = []  # 初始化最终文本块列表
        # 获取要使用的适当分隔符
        separator = separators[-1]  # 默认使用最后一个分隔符
        separator_rule = 'after'  # 默认分隔符规则为 'after'
        new_separators = []  # 初始化新的分隔符列表
        for i, _s in enumerate(separators):  # 遍历分隔符列表
            _separator = _s if self._is_separator_regex else re.escape(_s)  # 如果分隔符不是正则表达式，则进行转义
            separator_rule = self.separator_rule[_s]  # 获取分隔符规则
            if _s == '':  # 如果分隔符为空字符串
                separator = _s  # 设置分隔符为空字符串
                break  # 跳出循环
            if re.search(_separator, text):  # 如果在文本中找到分隔符
                separator = _s  # 设置分隔符为当前分隔符
                new_separators = separators[i + 1:]  # 更新新的分隔符列表，移除已使用的分隔符及其之前的分隔符
                break  # 跳出循环
        _separator = separator if self._is_separator_regex else re.escape(separator)  # 再次处理分隔符
        splits = _split_text_with_regex(text, _separator, self._keep_separator, separator_rule)  # 使用正则表达式分割文本
        # 现在合并内容，递归分割更长的文本。
        _good_splits = []  # 初始化好的文本块列表
        _separator = '' if self._keep_separator else separator  # 设置合并文本块时的分隔符，如果保留分隔符则为空字符串，否则为当前分隔符
        for s in splits:  # 遍历分割后的文本块
            if self._length_function(s) < self._chunk_size:  # 如果文本块长度小于 chunk_size
                _good_splits.append(s)  # 添加到好的文本块列表
            else:  # 如果文本块长度大于等于 chunk_size
                if _good_splits:  # 如果好的文本块列表不为空
                    merged_text = self._merge_splits(_good_splits, _separator)  # 合并好的文本块
                    final_chunks.extend(merged_text)  # 添加到最终文本块列表
                    _good_splits = []  # 清空好的文本块列表
                if not new_separators:  # 如果没有新的分隔符
                    final_chunks.append(s)  # 将当前文本块添加到最终文本块列表
                else:  # 如果有新的分隔符
                    other_info = self._split_text(s, new_separators)  # 递归分割文本块
                    final_chunks.extend(other_info)  # 添加到最终文本块列表
        if _good_splits:  # 如果好的文本块列表不为空
            merged_text = self._merge_splits(_good_splits, _separator)  # 合并好的文本块
            final_chunks.extend(merged_text)  # 添加到最终文本块列表
        return final_chunks  # 返回最终文本块列表

    def split_text(self, text: str) -> List[str]:
        """分割文本.
        使用预定义的分隔符列表分割文本。
        Args:
            text (str): 要分割的文本。
        Returns:
            List[str]: 分割后的文本块列表。
        """
        return self._split_text(text, self._separators)  # 调用 _split_text 方法进行文本分割

    def create_documents(self, texts: List[str], metadatas: Optional[List[dict]] = None) -> List[Document]:
        """从文本列表创建文档列表.
        将文本列表中的每个文本分割成多个文档块，并根据元数据信息，为每个文档块添加 bbox, chunk_type 等元数据信息。
        Args:
            texts (List[str]): 包含文本字符串的列表。
            metadatas (Optional[List[dict]]): 与文本列表对应的元数据列表，默认为 None。
        Returns:
            List[Document]: 创建的 Document 对象列表。
        """
        documents = []  # 初始化文档列表
        for i, text in enumerate(texts):  # 遍历文本列表
            index = -1  # 初始化索引
            # metadata = copy.deepcopy(_metadatas[i])
            indexes = metadatas[i].get('indexes', [])  # 获取索引信息
            pages = metadatas[i].get('pages', [])  # 获取页码信息
            types = metadatas[i].get('types', [])  # 获取类型信息
            bboxes = metadatas[i].get('bboxes', [])  # 获取 bbox 信息
            searcher = IntervalSearch(indexes)  # 创建 IntervalSearch 对象
            split_texts = self.split_text(text)  # 分割文本
            for chunk in split_texts:  # 遍历分割后的文本块
                new_metadata = copy.deepcopy(metadatas[i])  # 复制元数据
                if indexes and bboxes:  # 如果存在索引和 bbox 信息
                    index = text.find(chunk, index + 1)  # 查找文本块在原始文本中的索引
                    inter0 = [index, index + len(chunk) - 1]  # 计算文本块的间隔
                    norm_inter = searcher.find(inter0)  # 查找间隔对应的索引范围
                    new_metadata['chunk_bboxes'] = []  # 初始化 chunk_bboxes 列表
                    for j in range(norm_inter[0], norm_inter[1] + 1):  # 遍历索引范围
                        new_metadata['chunk_bboxes'].append({'page': pages[j], 'bbox': bboxes[j]})  # 添加 chunk_bbox 信息，包含页码和 bbox
                    c = Counter([types[j] for j in norm_inter])  # 统计索引范围内类型出现次数
                    chunk_type = c.most_common(1)[0][0]  # 获取出现次数最多的类型作为 chunk_type
                    new_metadata['chunk_type'] = chunk_type  # 设置 chunk_type
                    new_metadata['source'] = metadatas[i].get('source', '')  # 设置 source
                new_doc = Document(page_content=chunk, metadata=new_metadata)  # 创建 Document 对象
                documents.append(new_doc)  # 添加到文档列表
        return documents  # 返回文档列表

class ElemCharacterTextByTitleSplitter(RecursiveCharacterTextSplitter):
    """
    ElemCharacterTextByTitleSplitter 类，继承自 RecursiveCharacterTextSplitter，用于按标题分割文本。
    todo: 类描述待补充
    """

    def __init__(self, separators: Optional[List[str]] = None, separator_rule: Optional[List[str]] = None, is_separator_regex: bool = False, keep_separator: bool = True,
            **kwargs: Any, ) -> None:
        """创建 ElemCharacterTextByTitleSplitter 对象.
        Args:
            separators (Optional[List[str]]): 分隔符列表，用于文本分割，默认为 ['\n\n', '\n', ' ', '']。
            separator_rule (Optional[List[str]]): 分隔符规则列表，与 separators 列表对应，定义分隔符的使用方式，默认为 ['after', 'after', 'after', 'after']。
            is_separator_regex (bool): 是否将分隔符视为正则表达式，默认为 False。
            keep_separator (bool): 分割后的文本块是否保留分隔符，默认为 True。
            **kwargs (Any): 传递给 RecursiveCharacterTextSplitter 父类的其他参数。
        """
        super().__init__(separators=separators, keep_separator=keep_separator, **kwargs)
        self._separators = separators or ['\n\n', '\n', ' ', '']  # 设置分隔符列表，默认为 ['\n\n', '\n', ' ', '']
        self._separator_rule = separator_rule or ['after' for _ in range(4)]  # 设置分隔符规则列表，默认为 ['after', 'after', 'after', 'after']
        self.separator_rule = {one: self._separator_rule[index] for index, one in enumerate(separators)}  # 创建分隔符到规则的字典映射
        self._is_separator_regex = is_separator_regex  # 设置是否将分隔符视为正则表达式

    def split_documents(self, documents: Iterable[Document]) -> List[Document]:
        """分割文档列表.
        将文档列表中的每个文档分割成多个文档块。
        Args:
            documents (Iterable[Document]): 包含 Document 对象的 Iterable 对象。
        Returns:
            List[Document]: 分割后的 Document 对象列表。
        """
        texts, metadatas = [], []
        for doc in documents:  # 遍历文档列表
            texts.append(doc.page_content)  # 提取文档内容
            metadatas.append(doc.metadata)  # 提取文档元数据
        return self.create_documents(texts=texts, metadatas=metadatas)  # 调用 create_documents 方法创建分割后的文档

    def _split_text(self, text: str, separators: List[str]) -> List[str]:
        """按分隔符分割文本并返回文本块列表.
        递归地按分隔符分割文本，并返回分割后的文本块列表。
        Args:
            text (str): 要分割的文本。
            separators (List[str]): 分隔符列表。
        Returns:
            List[str]: 分割后的文本块列表。
        """
        final_chunks = []
        # Get appropriate separator to use
        separator = separators[-1]
        separator_rule = 'after'
        new_separators = []
        for i, _s in enumerate(separators):
            _separator = _s if self._is_separator_regex else re.escape(_s)
            separator_rule = self.separator_rule[_s]
            if _s == '':
                separator = _s
                break
            if re.search(_separator, text):
                separator = _s
                new_separators = separators[i + 1:]
                break
        _separator = separator if self._is_separator_regex else re.escape(separator)
        splits = _split_text_with_regex(text, _separator, self._keep_separator, separator_rule)
        # Now go merging things, recursively splitting longer texts.
        _good_splits = []
        _separator = '' if self._keep_separator else separator
        for s in splits:
            if self._length_function(s) < self._chunk_size:
                _good_splits.append(s)
            else:
                if _good_splits:
                    merged_text = self._merge_splits(_good_splits, _separator)
                    final_chunks.extend(merged_text)
                    _good_splits = []
                if not new_separators:
                    final_chunks.append(s)
                else:
                    other_info = self._split_text(s, new_separators)
                    final_chunks.extend(other_info)
        if _good_splits:
            merged_text = self._merge_splits(_good_splits, _separator)
            final_chunks.extend(merged_text)
        return final_chunks

    def split_text(self, text: str) -> List[str]:
        """分割文本.
        此方法重写父类方法，直接返回包含原始文本的列表，不进行分割。
        实际分割逻辑在 create_documents 方法中实现。
        Args:
            text (str): 要分割的文本。
        Returns:
            List[str]: 包含原始文本的列表，不进行分割。
        """
        return [text]  # 直接返回包含原始文本的列表，不进行分割

    def create_documents(self, texts: List[str], metadatas: Optional[List[dict]] = None) -> List[Document]:
        """从文本列表创建文档列表.
        将文本列表中的每个文本根据标题进行分割，并根据元数据信息，为每个文档块添加 chunk_bboxes, chunk_type, parents 等元数据信息。
        Args:
            texts (List[str]): 包含文本字符串的列表。
            metadatas (Optional[List[dict]]): 与文本列表对应的元数据列表，默认为 None。
        Returns:
            List[Document]: 创建的 Document 对象列表。
        """
        documents = []  # 初始化文档列表
        for i, text in enumerate(texts):  # 遍历文本列表
            index = -1  # 初始化索引
            # metadata = copy.deepcopy(_metadatas[i])
            indexes = metadatas[i].get('indexes', [])  # 获取索引信息
            pages = metadatas[i].get('pages', [])  # 获取页码信息
            types = metadatas[i].get('types', [])  # 获取类型信息
            bboxes = metadatas[i].get('bboxes', [])  # 获取 bbox 信息
            title_indexes = []  # 初始化标题索引列表
            title_contents = []  # 初始化标题内容列表
            level = metadatas[i].get('levels', [])  # 获取层级信息
            current_titles = []  # 初始化当前标题列表

            def find_content_titles():
                """查找内容标题.
                返回当前标题列表。
                """
                return current_titles  # 返回当前标题列表

            def add_current_title(new_title: dict, current_titles: List[dict]):
                """添加当前标题.
                将新的标题信息添加到当前标题列表中，并维护标题层级关系。
                Args:
                    new_title (dict): 新的标题信息字典，包含 level, title, id, page 等键。
                    current_titles (List[dict]): 当前标题列表。
                Returns:
                    List[dict]: 更新后的当前标题列表。
                """
                insert_index = -1  # 默认为-1，表示没有找到插入位置
                for i in range(len(current_titles)):  # 遍历当前标题列表
                    if current_titles[i]['level'] <= new_title['level']:  # 查找插入位置，根据标题层级判断
                        insert_index = i  # 找到插入位置
                        break
                parent_title = []  # 初始化父标题列表
                if insert_index == -1:  # 如果没有找到插入位置，表示新标题层级最高
                    parent_title = current_titles  # 父标题为当前所有标题
                else:  # 如果找到插入位置
                    parent_title = current_titles[0:insert_index]  # 父标题为插入位置之前的标题
                parent_id = [p["id"] for p in parent_title]  # 获取父标题 ID 列表
                new_title["parents"] = parent_id  # 设置新标题的 parents 属性为父标题 ID 列表
                if insert_index == -1:  # 如果没有找到插入位置，添加到列表末尾
                    current_titles.append(new_title)  # 添加到当前标题列表末尾
                else:  # 如果找到插入位置，插入到指定位置
                    current_titles.insert(insert_index, new_title)  # 插入到指定位置
                    current_titles = current_titles[:insert_index + 1]  # 截断当前标题列表，保持层级关系
                return current_titles  # 返回更新后的当前标题列表

            new_doc = None  # 初始化新文档对象
            t_type = "Title"  # 初始化文档类型为 "Title"
            new_metadata = {}  # 初始化新元数据字典
            merge_content = ""  # 初始化合并内容字符串
            chunk_end_flag = False  # 初始化分块结束标志
            for index, type_content in enumerate(types):  # 遍历类型列表和内容
                current_page = pages[index]  # 获取当前页码
                id = str(uuid.uuid4())  # 生成 UUID 作为文档 ID
                if type_content != "title" and t_type != "title" and not chunk_end_flag:  # 如果当前内容不是标题且上一个内容也不是标题且分块未结束
                    merge_content += text[indexes[index][0]:indexes[index][1] + 1]  # 合并内容
                    if "chunk_bboxes" not in new_metadata:  # 如果 chunk_bboxes 不在元数据中
                        new_metadata = {"source": metadatas[i].get('source', ''), "id": id, 'chunk_bboxes': [], 'page': current_page, "content_type": type_content,
                                        "level": level[index]}  # 初始化元数据
                else:  # 如果当前内容是标题或上一个内容是标题或分块已结束
                    merge_content = text[indexes[index][0]:indexes[index][1] + 1]  # 更新合并内容为当前内容
                    new_metadata = {"source": metadatas[i].get('source', ''), "id": id, 'chunk_bboxes': [], 'page': current_page, "content_type": type_content,
                                    "level": level[index]}  # 初始化元数据
                if self._chunk_size <= len(merge_content) + 50:  # 如果合并内容长度超过 chunk_size 阈值
                    chunk_end_flag = True  # 设置分块结束标志
                new_metadata['chunk_bboxes'].append({'page': pages[index], 'bbox': bboxes[index]})  # 添加 chunk_bbox 信息
                chunk_type = type_content  # 获取分块类型
                new_metadata['chunk_type'] = chunk_type  # 设置 chunk_type
                if type_content != 'title':  # 如果当前内容不是标题
                    parent_title = find_content_titles()  # 查找父标题
                    new_metadata['parents'] = [t_i["id"] for t_i in parent_title]  # 设置 parents 属性为父标题 ID 列表
                else:  # 如果当前内容是标题
                    current_titles = add_current_title({"level": level[index], "title": text[indexes[index][0]:indexes[index][1] + 1], "id": id, "page": current_page},
                        current_titles)  # 添加当前标题到当前标题列表
                    parent_title = find_content_titles()  # 查找父标题
                    if len(parent_title) >= 1:  # 如果父标题列表长度大于等于 1
                        parent_title = parent_title[:-1]  # 移除最后一个父标题，保持层级关系
                    new_metadata['parents'] = [t_i["id"] for t_i in parent_title]  # 设置 parents 属性为父标题 ID 列表
                source = new_metadata["source"]  # 获取 source
                file_name = source.split(".")[0]  # 提取文件名
                new_doc = Document(page_content=f"{file_name}\n{merge_content}", metadata=new_metadata)  # 创建 Document 对象，page_content 包含文件名和合并内容
                if type_content != "title" and t_type != "title" and not chunk_end_flag:  # 如果当前内容不是标题且上一个内容也不是标题且分块未结束
                    pass  # 什么也不做，继续合并内容
                else:  # 否则，添加到文档列表
                    chunk_end_flag = False  # 重置分块结束标志
                    documents.append(new_doc)  # 添加到文档列表
                t_type = type_content  # 更新文档类型为当前内容类型
        return documents  # 返回文档列表



import os
import gradio as gr
import pandas as pd
from pathlib import Path
from gradio import components
from bisheng_langchain.rag.scoring.ragas_score import RagScore


save_folder = '/home/public/rag_benchmark_v1.0/rag_score_demo'
if not os.path.exists(save_folder):
    os.makedirs(save_folder)


def rag_evaluate(excel_file, metric='answer_recall_bisheng', batch_size=5):
    excel_path = excel_file.name
    df = pd.read_excel(excel_path)
    if '问题类型' not in df.columns:
        df['问题类型'] = len(df['问题'].tolist()) * ['QA']
    df.to_excel(excel_path, index=False)
    params = {
        'excel_path': excel_path,
        'save_path': save_folder,
        'question_column': '问题',
        'query_type_column': '问题类型',
        'gt_column': '人工标注',
        'answer_column': '模型回答',
        'metrics': [metric],
        'batch_size': int(batch_size),
    }
    rag_score = RagScore(**params)
    rag_score.score()

    output_path = Path(save_folder) / f"{Path(excel_path).stem}_score.xlsx"
    return str(output_path)


if __name__ == '__main__':
     title = """毕昇QA问答自动评估系统"""
     with gr.Blocks() as demo:
        gr.Markdown(title)
        
        with gr.Row():
            with gr.Column(scale=2):
                with gr.Row():
                    eval_intput_file = gr.components.File(label='FlowFile')
                    with gr.Column():
                        metric_options = ["answer_recall_bisheng", "answer_correctness_bisheng"]
                        metric = gr.components.Dropdown(label="评估方法", choices=metric_options, default=metric_options[0], interactive=True)
                        # metric = gr.Textbox(label='评估方法', value='answer_recall_bisheng', interactive=True, lines=2)
                        batch_size = gr.Textbox(label='批评估大小', value=5, interactive=True, lines=2)
                btn0 = gr.Button('Run Evaluation')
            eval_out_file = gr.components.File(label='FlowFile')

        btn0.click(fn=rag_evaluate, inputs=[eval_intput_file, metric, batch_size], outputs=[eval_out_file])
        demo.queue().launch(share=False, inbrowser=True, server_name="0.0.0.0", server_port=8218)


import json
from typing import Any, Dict, List, Optional
from venv import logger

from bisheng_langchain.chains import LoaderOutputChain
from langchain.callbacks.manager import AsyncCallbackManagerForChainRun, CallbackManagerForChainRun
from langchain.chains.base import Chain
from pydantic import BaseModel, Extra

_TEXT_COLOR_MAPPING = {
    'blue': '36;1',
    'yellow': '33;1',
    'pink': '38;5;200',
    'green': '32;1',
    'red': '31;1',
}


def get_color_mapping(
    items: List[str], excluded_colors: Optional[List] = None
) -> Dict[str, str]:
    """Get mapping for items to a support color."""
    colors = list(_TEXT_COLOR_MAPPING.keys())
    if excluded_colors is not None:
        colors = [c for c in colors if c not in excluded_colors]
    color_mapping = {item: colors[i % len(colors)] for i, item in enumerate(items)}
    return color_mapping


class Output(BaseModel):
    """Output组件，用来控制输出"""

    @classmethod
    def initialize(cls, file_path: str = None):
        return file_path if file_path else ''


class Report(Chain):
    # ```
    # chain Dict:
    #    object: langchain_object
    #    node_id: object_key prefix
    #    input: triger query
    # variables Dict:
    #    variable_name: name
    #    variable_value: value
    # `
    chains: Optional[List[Dict]]
    variables: Optional[List[Dict]]
    report_name: str
    stop_flag: bool = False

    input_key: str = 'report_name'  #: :meta private:
    output_key: str = 'text'  #: :meta private:

    class Config:
        """Configuration for this pydantic object."""
        extra = Extra.forbid
        arbitrary_types_allowed = True

    @property
    def input_keys(self) -> List[str]:
        """Expect input key.
        :meta private:
        """
        return [self.input_key]

    @property
    def output_keys(self) -> List[str]:
        """Return output key.
        :meta private:
        """
        return [self.output_key]

    def validate_chains(cls, values: Dict) -> Dict:
        """Validate chains."""
        if values.get('chains'):
            for chain in values['chains']:
                chain_output_keys = chain['object'].output_keys
                if len(chain_output_keys) != 1:
                    raise ValueError(
                        'Chain used in Report should all have one output, got '
                        f"{chain['object']} with {len(chain_output_keys)} outputs."
                    )
            return values

    def func_call(self,
                  inputs: Dict[str, Any],
                  outputs: Dict[str, Any],
                  intermedia_stop: list,
                  chain: Chain,
                  node_id: str,
                  run_manager: Optional[CallbackManagerForChainRun] = None,):
        question = list(inputs.values())[0]
        _run_manager = run_manager or CallbackManagerForChainRun.get_noop_manager()

        if isinstance(chain, LoaderOutputChain):
            question = 'Get' + ','.join(question)
        _run_manager.on_text(text='', log='', type='start', category='question')
        _run_manager.on_text(text='', log=question, type='end', category='question')
        _run_manager.on_text(text='', log='', type='start', category='answer')
        message_reply = {'log': question, 'category': 'question'}
        intermedia_stop.append(message_reply)

        chain_outputs = chain(inputs, callbacks=_run_manager.get_child())
        result = (chain_outputs.get(chain.output_keys[0])
                  if isinstance(chain_outputs, dict) else chain_outputs)
        if isinstance(chain, LoaderOutputChain):
            for schema in inputs.values():
                result = json.loads(result)
                for key in schema:
                    if result.get(key):
                        result_str = ('；'.join([str(x) for x in result.get(key)])
                                      if isinstance(result.get(key), list)
                                      else result.get(key))
                        outputs.update({node_id+'_'+key: result_str})
            result = json.dumps(result, ensure_ascii=False)
        else:
            outputs.update({node_id: result})
        message_reply = {'log': result, 'category': 'answer'}
        intermedia_stop.append(message_reply)
        _run_manager.on_text(text='', log=result, type='end', category='answer')

    async def func_acall(self,
                         inputs: Dict[str, Any],
                         outputs: Dict[str, Any],
                         intermedia_stop: list,
                         chain: Chain,
                         node_id: str,
                         run_manager: Optional[AsyncCallbackManagerForChainRun] = None,):
        question = list(inputs.values())[0]
        _run_manager = run_manager or AsyncCallbackManagerForChainRun.get_noop_manager()

        if isinstance(chain, LoaderOutputChain):
            question = 'Get' + ','.join(question)
        await _run_manager.on_text(text='', log='', type='start', category='question')
        await _run_manager.on_text(text='', log=question, type='end', category='question')
        await _run_manager.on_text(text='', log='', type='start', category='answer')
        message_reply = {'log': question, 'category': 'question'}
        intermedia_stop.append(message_reply)

        # process
        try:
            chain_outputs = await chain.arun(inputs, callbacks=_run_manager.get_child())
        except Exception as e:
            logger.exception(e)
            await _run_manager.on_text(text='', log=str(e), type='stream', category='processing')
            try:
                chain_outputs = chain(inputs)
            except Exception as e2:
                logger.exception(e2)
                await _run_manager.on_text(text='', log=str(e2), type='stream', category='processing')
                chain_outputs = ''

        result = (chain_outputs.get(chain.output_keys[0])
                  if isinstance(chain_outputs, dict) else chain_outputs)
        if isinstance(chain, LoaderOutputChain):
            for schema in inputs.values():
                result = json.loads(result)
                for key in schema:
                    if result.get(key):
                        result_str = ('；'.join([str(x) for x in result.get(key)])
                                      if isinstance(result.get(key), list)
                                      else result.get(key))
                        outputs.update({node_id+'_'+key: result_str})
            result = json.dumps(result, ensure_ascii=False)
        else:
            outputs.update({node_id: result})
        message_reply = {'log': result, 'category': 'answer'}
        intermedia_stop.append(message_reply)
        await _run_manager.on_text(text='', log=result, type='end', category='answer')

    def _call(
        self,
        inputs: Dict[str, Any],
        run_manager: Optional[CallbackManagerForChainRun] = None,
        verbose: Optional[bool] = None,
    ) -> Dict[str, str]:
        intermedia_steps = []
        outputs = {}
        self.stop_flag = False
        # variables
        if self.variables and self.variables[0]:
            for variable in self.variables:
                variable_kv = variable['input']
                for k, v in variable_kv.items():
                    outputs.update({variable['node_id']+'_'+k: v})

        if self.chains:
            for i, chain in enumerate(self.chains):
                if 'node_id' not in chain:
                    logger.info(f"report_skip_nonsence_chain chain={chain['object']}")
                    continue
                if not isinstance(chain['object'], Chain):
                    raise TypeError(
                        f"{chain['object']} not be runnable Chain object"
                    )
                if isinstance(chain['object'], LoaderOutputChain):
                    # loaderchain questions use new parse
                    self.func_call(chain['input'], outputs, intermedia_steps,
                                   chain['object'], chain['node_id'], run_manager)
                    continue

                preset_question = chain['input']
                for k, v in preset_question.items():
                    # log print
                    if isinstance(v, str):
                        self.func_call(preset_question, outputs, intermedia_steps,
                                       chain['object'], chain['node_id']+'_'+v, run_manager)
                    else:
                        for question in v:
                            question_dict = {k: question}
                            self.func_call(question_dict, outputs, intermedia_steps,
                                           chain['object'], chain['node_id']+'_'+question,
                                           run_manager)

        return {self.output_key: outputs, self.input_key: self.report_name,
                'intermediate_steps': intermedia_steps}

    async def _acall(
        self,
        inputs: Dict[str, Any],
        run_manager: Optional[AsyncCallbackManagerForChainRun] = None,
        verbose: Optional[bool] = None,
    ) -> Dict[str, Any]:
        intermedia_steps = []
        outputs = {}
        await run_manager.on_text(text='', log='', type='end', category='processing')  # end father start
        self.stop_flag = False
        # variables
        if self.variables and self.variables[0]:
            for variable in self.variables:
                variable_kv = variable['input']
                for k, v in variable_kv.items():
                    outputs.update({variable['node_id']+'_'+k: v})

        # functions
        if self.chains:
            for i, chain in enumerate(self.chains):
                if 'node_id' not in chain:
                    logger.info(f"report_skip_nonsence_chain chain={chain['object']}")
                    continue
                if self.stop_flag:
                    break
                if not isinstance(chain['object'], Chain):
                    raise TypeError(
                        f"{chain['object']} not be runnable Chain object"
                    )
                if isinstance(chain['object'], LoaderOutputChain):
                    # loaderchain questions use new parse
                    await self.func_acall(chain['input'], outputs, intermedia_steps,
                                          chain['object'], chain['node_id'], run_manager)
                    continue
                # normal chain
                preset_question = chain['input']
                for k, v in preset_question.items():
                    if isinstance(v, str):
                        await self.func_acall(preset_question, outputs, intermedia_steps,
                                              chain['object'], chain['node_id']+'_'+v, run_manager)
                    else:
                        for question in v:
                            question_dict = {k: question}
                            await self.func_acall(question_dict, outputs, intermedia_steps,
                                                  chain['object'], chain['node_id']+'_'+question,
                                                  run_manager)

        # keep whole process paired
        await run_manager.on_text(text='', log='', type='start', category='processing')
        return {self.output_key: outputs, self.input_key: self.report_name,
                'intermediate_steps': intermedia_steps}

    def stop(self):
        self.stop_flag = True

    def stop_status(self):
        return self.stop_flag



import os
import openai
import httpx
import time
import json
import shutil
import pandas as pd
from tqdm import tqdm
from collections import defaultdict
from openai import OpenAI
thread_ids = set()


client = OpenAI(api_key=os.environ.get('OPENAI_API_KEY', ''), 
                http_client=httpx.Client(proxies=os.environ.get('OPENAI_PROXY', '')))


def upload_files(data_folder, excel_file, save_excel_file):
    df = pd.read_excel(excel_file)
    all_questions_info = list()
    # 遍历每一行
    for index, row in df.iterrows():
        # 遍历每一列
        question_info = dict()
        for column in df.columns:
            value = row[column]
            question_info[column] = value
        all_questions_info.append(question_info)
    
    file2collection = dict()
    for questions_info in tqdm(all_questions_info):
        file_name = questions_info['文件名']
        if file_name not in file2collection:
            pdf_file = os.path.join(data_folder, file_name)
            response = client.files.create(
                file=open(pdf_file, "rb"),
                purpose="assistants"
            )
            file_id = response.id
            file2collection[file_name] = file_id
            questions_info['openai_assistant_file_id'] = file_id
        else:
            file_id = file2collection[file_name]
            questions_info['openai_assistant_file_id'] = file_id

    # save excel
    df = pd.DataFrame(all_questions_info)
    df.to_excel(save_excel_file, index=False)


def question_answer(excel_file):
    df = pd.read_excel(excel_file)
    all_questions_info = list()
    # 遍历每一行
    for index, row in df.iterrows():
        # 遍历每一列
        question_info = dict()
        for column in df.columns:
            value = row[column]
            question_info[column] = value
        all_questions_info.append(question_info)
    
    file2assistant = dict()
    for questions_info in tqdm(all_questions_info):
        question = questions_info['问题']
        file_id = questions_info['openai_assistant_file_id']
        if file_id not in file2assistant:
            assistant = client.beta.assistants.create(
                name='文档问答系统',
                instructions="根据上传的文档，回答相关的问题.",
                model="gpt-4-1106-preview",
                tools=[{"type": "retrieval"}],
                file_ids=[file_id]
            )
            file2assistant[file_id] = assistant
        else:
            assistant = file2assistant[file_id]

        thread = client.beta.threads.create()
        while (thread.id in thread_ids):
            thread = client.beta.threads.create()
        thread_ids.add(thread.id)
        
        message = client.beta.threads.messages.create(
            thread_id=thread.id,
            role="user",
            content=question
        )
        run = client.beta.threads.runs.create(
            thread_id=thread.id,
            assistant_id=assistant.id,
        )
        run = client.beta.threads.runs.retrieve(
            thread_id=thread.id,
            run_id=run.id
        )
        while run.status != 'completed':
            run = client.beta.threads.runs.retrieve(
                thread_id=thread.id,
                run_id=run.id
            )
            print(run.status)
            time.sleep(1)
    
        # print(run)
        messages = client.beta.threads.messages.list(thread_id=thread.id)
        ans = messages.data[0].content[0].text.value
        questions_info['openai_assistant_answer'] = ans
        print('query:', question)
        print('pred:', ans)
        print('--------------------------------------')
    
    # save excel
    df = pd.DataFrame(all_questions_info)
    df.to_excel(excel_file, index=False)


if __name__ == '__main__':
    data_folder = '/home/public/rag_benchmark_v1.0/rag_benchmark_processed'
    excel_file = '../data/questions_info_with_answer_sample.xlsx'
    save_excel_file = '../data/questions_info_with_openai_assistant_file_id.xlsx'
    # upload_files(data_folder, excel_file, save_excel_file)
    question_answer(save_excel_file)


# import base64
import copy
import requests
import base64
from typing import Any, Iterator, List, Mapping, Optional, Union


class OCRClient(object):
    def __init__(self,
                 api_base_url: Optional[str] = None):
        # http://192.168.106.12:36001/v2/idp/idp_app/infer
        self.ep = api_base_url
        self.client = requests.Session()
        self.timeout = 10000
        self.params = {
            'sort_filter_boxes': True,
            'enable_huarong_box_adjust': True,
            'support_long_image_segment': True,
            'rotateupright': False,
        }

        self.scene_mapping = {
            'doc': {
                'det': 'general_text_det_mrcnn_v1.0',
                'recog': 'transformer-v2.8-gamma-faster'
            },
            'form': {
                'det': 'mrcnn-v5.1',
                'recog': 'transformer-v2.8-gamma-faster'
            },
            'hand': {
                'det': 'mrcnn-v5.1',
                'recog': 'transformer-hand-v1.16-faster'
            }
        }

    def predict(self, inp):
        scene = inp.pop('scene', 'form')
        b64_image = inp.pop('b64_image')
        params = copy.deepcopy(self.params)
        params.update(self.scene_mapping[scene])
        params.update(inp)

        req_data = {'data': [b64_image], 'param': params}

        try:
            r = self.client.post(url=self.ep,
                             json=req_data,
                             timeout=self.timeout)
            return r.json()
        except Exception as e:
            return {'status_code': 400, 'status_message': str(e)}




from langchain_core.prompts.chat import (
    ChatPromptTemplate,
    HumanMessagePromptTemplate,
    SystemMessagePromptTemplate,
)


SEED_QUESTION_SYSTEM = SystemMessagePromptTemplate.from_template(
    """\
您的任务是遵循以下规则从给定的上下文中提出一个问题，规则如下：

    1. 即使在没有给定上下文的情况下，问题也应该对人类有意义。
    2. 应该可以从给定上下文中完全回答问题。
    3. 问题应该来自包含重要信息的上下文部分。它也可以来自表格、段落、或者代码等。
    4. 回答问题时不应包含任何链接。
    5. 问题的难度应该是中等的。
    6. 问题必须是合理的，并且必须能被人理解和回答。
    7. 不要在问题中使用“提供的上下文”等短语。
    8. 避免使用可以分解成多个问题的“和”字样来构建问题。
    9. 如果上下文是中文，那么问题也应该是中文的。

Examples:
context:武汉达梦数据库股份有限公司 招股说明书 （申报稿） 1-1-226 表中作出恰当列报。 2、研发费用 2021年度、 2020年度、 2019 年度，达梦数据 研发费用金额分别 为11,786.99 万元、 9,660.26 万元、 6,255.86万元， 各年度研发费用占营 业收入的比例分别为 15.86 % 、 21.46 %、20.74 %。 由于研发投入金额及其占当期 营业收入的比例是 达梦数据 的关键 指标之一，可能存在因为核算不准 确而导致的错报风险。因此， 中天 运会计师 将研发费用的归集和核算 确定为关键审计事项。 针对研发费用的真实性与准确性，会计师执行的 重要审计程序主要包括： （1）了解与研发费用相关的关键内部控制，评价 这些控制的设计，确定其是否得到执行，并对相关内 部控制的运行有效性进行测试； （2）获取研发项目立项、审批资料，抽查重要研 发项目的过程文档，判断研发项目的真实性； （3）获取研发费用按项目、性质分类明细表，分
question:达梦2021年的研发费用占营业收入的比例是多少？

context:武汉达梦数据库股份有限公司 招股说明书 （申报稿） 1-1-329 （2）存货周转率 公司与同行业可比公司存货周转率对比情况如下： 公司简称 2021年度 2020年度 2019年度 中望软件 6.93 5.62 10.66 星环科技 3.38 3.21 2.24 金山办公 212.60 175.46 162.91 平均值 74.30 61.43 58.60 本公司 1.13 0.57 0.87 数据来源：可比公司招股说明书、定期报告。 报告期各期， 公司存货周转率显著低于同行业可比公司存货周转率平均水平， 主要是因为公司将未验收的数据及行业解决方案项目所发生的累 计成本均作为 存货核算。报告期各期末，公司存在 “湖北省司法行政数据中心项目 ”、“政法云 大数据中心基础设施服务及大数据中心软件采购 项目”等金额较大且实施周期较 长的数据及行业解决方案项目，导致年末存货金额较大。
question:达梦2021年的存货周转率相较于前一年有何变化？
"""  # noqa: E501
)


SEED_QUESTION_HUMAN = HumanMessagePromptTemplate.from_template(
"""
context:{context}
question:
"""
)


SEED_QUESTION_CHAT_PROMPT = ChatPromptTemplate.from_messages(
    [
        SEED_QUESTION_SYSTEM,
        SEED_QUESTION_HUMAN
    ]
)


SCORE_CONTEXT_SYSTEM = SystemMessagePromptTemplate.from_template(
"""Evaluate the provided context and assign a numerical score between 0 and 10 based on the following criteria:
1. Award a high score to context that thoroughly delves into and explains concepts.
2. Assign a lower score to context that contains excessive references, acknowledgments, external links, personal information, or other non-essential elements.

And you should only output the score.

Examples:
Context:
01-2022.04.30 贷方发生额共 计 2535.43 万元，户名；X 贸易有限公司；\n③根据用款企业提供的增值税纳税申报表来看，2021 年度用款企业年累计开票额为\n7826.48 万元，年累计应纳税合计 95.32 万元，年累计已纳税额 86.23 万元；截止至 2022 年 3 月，用款企业累计开票额为 1986.54 万元，累计应纳税合计19.54 万元，累计已纳税额\n20.23 万元。\n根据核算用款企业的银行流水及企业会计记账系统，剔除借款人往来转账款，估算用款 企业年营业额约在 6000 万元左右(纳税申报营业额)，全部营业收入约 20000 万元左右，借 款人所在 X 贸易有限公司综合毛
利润率约为 35%，净利润约 20%左右。\n\n| 资产种类 | 坐落 | 产权人 | 建筑面积 | 现价值 | 贷款余额 | 资产净值 |\n| --- | --- | --- | --- | --- | --- | --- |\n| 房产 | HN 省 YY 市 PP 小区 5#2-101 | A | 240.20 | 365.23 万 | 165.
Score: 4

Context:
认缴出资额 200 万元 实缴出资额 200 万元 持股比例 20% |\n| 企业所属商圈 | 无 | 是否为已准入商圈 | 是□ 否 ☑ |\n(1) 企业经营历史及现状说明\nX 贸易有限公司 (下称“用款企业”) 注册成立于 2015 年 11 月，统一社会信用代码1234567890ACBDEFGH，法定代表人 A，公司注册地址位于 M 市 N 区 JF 路 20 号 NJ 大厦 18 楼1807 室，实际办公地址位于 M 市 N 区 K 广场 C 座 19 楼 1901、1906、1908、1910、1912、1914，办公面积为 880.51 ㎡，经营场所为用款企业租赁房产，租赁期限，现阶段年租金 73 万余元。\n用款企业是著名品牌“XYZ”的运营公司，是
以经营短袜、连裤袜、 内衣、家居服、配饰为主要品类的亲体织物公司，致力于为年轻消费群体提供“一站式”多品类亲体织物购物 体验。 作为织物文化的传播者和输出者，用款企业秉承一贯的高品质与原创精神，依托中国 研发团队，创领多项核心技术，不断建立并升级健康织物行业标准，目前拥有实用新型专利 6 项，发明专利 1 项，注册商标 30 余个，为品牌的商标保护构建了全面的商标防御体系。\n“XYZ”品牌创立于 2006 年，于 2009 年正式进入中国市场，在成立 10 年的时间里，在全国共有 400 余家店面，运营主要有以下三种模式：\n①直营模式：目前用款企业
管控的直营店有 100 家左右，其中在 M 地区共有 9 家直营店，分别为 Y1 店、Y2 店、Y3 店、Y4 店、Y5 店、Y6 店、Y7 店、Y8 店、Y9 店。经查看用款企业相关财务系统并截屏 ，用款企业 2021 年度 、2022 年 1-4 月直营店营业收入合计分别为7623.45 万元、1987.23 万元，M 地区 9 家直营店收入合计分别为 1238.67 万元、302.54 万元。根据数据测算直营部分毛利润率65%。
Score: 7
"""  # noqa: E501
)


SCORE_CONTEXT_HUMAN = HumanMessagePromptTemplate.from_template(
"""
Context:
{context}
Score:
"""  # noqa: E501
)


SCORE_CONTEXT_CHAT_PROMPT = ChatPromptTemplate.from_messages(
    [
        SCORE_CONTEXT_SYSTEM,
        SCORE_CONTEXT_HUMAN
    ]
)


FILTER_QUESTION_SYSTEM = SystemMessagePromptTemplate.from_template(
    """\
Determine if the given question can be clearly understood even when presented without any additional context. Specify reason and verdict is a valid json format.

Examples:
question: What is the discovery about space?
{{
    "reason":"The question is too vague and does not specify which discovery about space it is referring to."
    "verdit":"No"
}}

question: What caused the Great Depression?
{{
    "reason":"The question is specific and refers to a well-known historical economic event, making it clear and answerable.",
    "verdict":"Yes"
}}

question: What is the keyword that best describes the paper's focus in natural language understanding tasks?
{{
    "reason": "The question mentions a 'paper' in it without referring it's name which makes it unclear without it",
    "verdict": "No"
}}

question: Who wrote 'Romeo and Juliet'?
{{
    "reason": "The question is clear and refers to a specific work by name therefore it is clear",
    "verdict": "Yes"
}}

question: What did the study mention?
{{
    "reason": "The question is vague and does not specify which study it is referring to",
    "verdict": "No"
}}

question: What is the focus of the REPLUG paper?
{{
    "reason": "The question refers to a specific work by it's name hence can be understood", 
    "verdict": "Yes"
}}
"""  # noqa: E501
)


FILTER_QUESTION_HUMAN = HumanMessagePromptTemplate.from_template(
    """\
question:{question}
"""  # noqa: E501
)


FILTER_QUESTION_CHAT_PROMPT = ChatPromptTemplate.from_messages(
    [
        FILTER_QUESTION_SYSTEM,
        FILTER_QUESTION_HUMAN
    ]
)


ANSWER_FORMULATE = HumanMessagePromptTemplate.from_template(
    """\
Answer the question using the information from the given context. 

context:{context}

question:{question}
answer:
"""  # noqa: E501
)



# flake8: noqa  # 禁用 flake8 代码检查
"""使用语义分割器加载 PDF."""
import base64  # 导入 base64 模块，用于 base64 编码和解码
import logging  # 导入 logging 模块，用于日志记录
import os  # 导入 os 模块，用于操作系统相关的功能
from typing import List  # 导入 List 类型注解
import requests  # 导入 requests 库，用于发送 HTTP 请求
from langchain.docstore.document import Document  # 导入 Langchain 的 Document 类，用于表示文档
from langchain.document_loaders.pdf import BasePDFLoader  # 导入 Langchain 的 BasePDFLoader 类，作为 PDF 加载器的基类
logger = logging.getLogger(__name__)  # 获取名为 __name__ 的 logger 实例

def merge_partitions(partitions):
    """合并分区数据.
    将非结构化文档解析 API 返回的分区数据合并成一个文本内容和一个元数据字典。
    Args:
        partitions: 分区数据列表，每个元素是一个字典，包含 'type' (分区类型), 'text' (分区文本), 'metadata' (元数据)。
    Returns:
        Tuple[str, Dict]: 合并后的文本内容和元数据字典。
    """
    text_elem_sep = '\n'  # 文本元素分隔符
    doc_content = []  # 初始化文档内容列表
    is_first_elem = True  # 标记是否为第一个元素
    last_label = ''  # 上一个分区的类型标签
    prev_length = 0  # 之前内容的长度
    metadata = dict(bboxes=[], pages=[], indexes=[], types=[], levels=[])  # 初始化元数据字典，用于存储 bbox, 页码, 索引, 类型, 层级信息
    for part in partitions:  # 遍历分区数据列表
        label, text = part['type'], part['text']  # 获取分区类型和文本
        extra_data = part['metadata']['extra_data']  # 获取额外元数据
        offset = 0  # 初始化偏移量
        if is_first_elem:  # 如果是第一个元素
            if label == 'Title':  # 如果是标题类型
                f_text = text + '\n'  # 标题后添加换行符
            else:  # 如果不是标题类型
                f_text = text  # 直接使用文本
            doc_content.append(f_text)  # 添加到文档内容列表
            is_first_elem = False  # 设置为非第一个元素
        else:  # 如果不是第一个元素
            if last_label == 'Title' and label == 'Title':  # 如果上一个和当前都是标题类型
                doc_content.append('\n' + text)  # 添加单换行符和文本
                offset = 1  # 设置偏移量为 1
            elif label == 'Title':  # 如果当前是标题类型
                doc_content.append('\n\n' + text)  # 添加双换行符和文本
                offset = 2  # 设置偏移量为 2
            elif label == 'Table':  # 如果当前是表格类型
                doc_content.append('\n\n' + text)  # 添加双换行符和文本
                offset = 2  # 设置偏移量为 2
            else:  # 如果是其他类型
                if last_label == 'Table':  # 如果上一个是表格类型
                    doc_content.append(text_elem_sep * 2 + text)  # 添加双倍元素分隔符和文本
                    offset = 2  # 设置偏移量为 2
                else:  # 如果上一个不是表格类型
                    doc_content.append(text_elem_sep + text)  # 添加元素分隔符和文本
                    offset = 1  # 设置偏移量为 1
        last_label = label  # 更新上一个分区的类型标签
        metadata['bboxes'].extend(list(map(lambda x: list(map(int, x)), extra_data['bboxes'])))  # 提取并添加 bbox 信息
        metadata['pages'].extend(extra_data['pages'])  # 添加页码信息
        metadata['types'].extend(extra_data['types'])  # 添加类型信息
        metadata['levels'].extend(extra_data['levels'])  # 添加层级信息
        indexes = extra_data['indexes']  # 获取索引信息
        up_indexes = [[s + prev_length + offset, e + prev_length + offset] for (s, e) in indexes]  # 更新索引信息，考虑之前内容的长度和偏移量
        metadata['indexes'].extend(up_indexes)  # 添加更新后的索引信息
        prev_length += len(doc_content[-1])  # 更新之前内容的长度
    content = ''.join(doc_content)  # 合并文档内容列表为字符串
    return content, metadata  # 返回合并后的内容和元数据

class ElemUnstructuredLoader(BasePDFLoader):
    """ElemUnstructuredLoader 类，继承自 BasePDFLoader，使用 unstructured API 加载 PDF 文档并进行语义分割。
    使用 pypdf 加载 PDF，并在字符级别进行分块。 虚拟版本
    加载器还在元数据中存储页码。
    """

    def __init__(self, file_name: str, file_path: str, unstructured_api_key: str = None, unstructured_api_url: str = None, start: int = 0, n: int = None, verbose: bool = False,
                 kwargs: dict = {}) -> None:
        """使用文件路径初始化 ElemUnstructuredLoader 对象.
        Args:
            file_name (str): 文件名。
            file_path (str): 文件路径。
            unstructured_api_key (str, optional): unstructured API 密钥，默认为 None。
            unstructured_api_url (str, optional): unstructured API URL，默认为 None。
            start (int): 分区起始页码，默认为 0。
            n (int): 分区页数，默认为 None，表示所有页。
            verbose (bool): 是否启用详细日志输出，默认为 False。
            kwargs (dict): 传递给 unstructured API 的额外参数字典，默认为空字典。
        """
        self.unstructured_api_url = unstructured_api_url  # 设置 unstructured API URL
        self.unstructured_api_key = unstructured_api_key  # 设置 unstructured API 密钥
        self.headers = {'Content-Type': 'application/json'}  # 设置 HTTP 请求头，指定 Content-Type 为 application/json
        self.file_name = file_name  # 设置文件名
        self.start = start  # 设置分区起始页码
        self.n = n  # 设置分区页数
        self.extra_kwargs = kwargs  # 设置额外参数
        self.partitions = None  # 初始化分区数据为 None
        super().__init__(file_path)  # 调用父类 BasePDFLoader 的初始化方法

    def load(self) -> List[Document]:
        """将给定路径加载为 Document 页面列表.
        调用 unstructured API 对 PDF 文件进行分区解析，并将解析结果转换为 Langchain Document 对象列表。
        Returns:
            List[Document]: 包含解析后的文档内容的 Document 对象列表，通常只包含一个 Document 对象。
        """
        b64_data = base64.b64encode(open(self.file_path, 'rb').read()).decode()  # 将文件内容 base64 编码
        parameters = {'start': self.start, 'n': self.n}  # 设置 API 请求参数，包含起始页码和页数
        parameters.update(self.extra_kwargs)  # 合并额外参数
        payload = dict(filename=os.path.basename(self.file_name), b64_data=[b64_data], mode='partition', parameters=parameters)  # 构建 API 请求 payload，包含文件名, base64 编码数据, 解析模式, 参数
        resp = requests.post(self.unstructured_api_url, headers=self.headers, json=payload)  # 发送 POST 请求到 unstructured API
        if resp.status_code != 200:  # 如果 API 响应状态码不是 200，表示请求失败
            raise Exception(f'文件分区 {os.path.basename(self.file_name)} 失败 resp={resp.text}')  # 抛出异常，提示文件分区失败
        resp = resp.json()  # 解析 JSON 响应
        if 200 != resp.get('status_code'):  # 如果 API 返回的状态码不是 200，表示解析服务内部错误
            logger.info(f'文件分区 {os.path.basename(self.file_name)} 错误 resp={resp}')  # 记录错误日志
            raise Exception(f'文件分区错误 {os.path.basename(self.file_name)} 错误 resp={resp}')  # 抛出异常，提示文件分区错误
        partitions = resp['partitions']  # 获取分区数据
        if partitions:  # 如果分区数据不为空
            logger.info(f'content_from_partitions')  # 记录日志，提示内容来自 partitions
            self.partitions = partitions  # 存储分区数据
            content, metadata = merge_partitions(partitions)  # 合并分区数据，获取文本内容和元数据
        elif resp.get('text'):  # 如果分区数据为空，但响应中包含 text 字段
            logger.info(f'content_from_text')  # 记录日志，提示内容来自 text 字段
            content = resp['text']  # 使用 text 字段内容
            metadata = {"bboxes": [], "pages": [], "indexes": [], "types": [], "levels": []}  # 初始化元数据为空字典
        else:  # 如果分区数据和 text 字段都为空
            logger.warning(f'content_is_empty resp={resp}')  # 记录警告日志，提示内容为空
            content = ''  # 设置内容为空字符串
            metadata = {}  # 设置元数据为空字典
        logger.info(f'unstruct_return code={resp.get("status_code")}')  # 记录 API 返回的状态码
        if resp.get('b64_pdf'):  # 如果响应中包含 b64_pdf 字段，表示返回了转换后的 PDF 文件 base64 编码
            with open(self.file_path, 'wb') as f:  # 将 base64 编码的 PDF 数据解码并保存到文件
                f.write(base64.b64decode(resp['b64_pdf']))  # base64 解码并写入文件
        metadata['source'] = self.file_name  # 设置元数据中的 source 字段为文件名
        doc = Document(page_content=content, metadata=metadata)  # 创建 Langchain Document 对象
        return [doc]  # 返回包含 Document 对象的列表

class ElemUnstructuredLoaderV0(BasePDFLoader):
    """ElemUnstructuredLoaderV0 类，继承自 BasePDFLoader，使用 unstructured API 加载各种文件类型并提取文本。
    根据文件格式自动选择合适的解析器，并支持 OCR。
    """

    def __init__(self, file_name: str, file_path: str, unstructured_api_key: str = None, unstructured_api_url: str = None, start: int = 0, n: int = None, verbose: bool = False,
                 kwargs: dict = {}) -> None:
        """使用文件路径初始化 ElemUnstructuredLoaderV0 对象.
        Args:
            file_name (str): 文件名。
            file_path (str): 文件路径。
            unstructured_api_key (str, optional): unstructured API 密钥，默认为 None。
            unstructured_api_url (str, optional): unstructured API URL，默认为 None。
            start (int): 分区起始页码，默认为 0 (text 模式下无效)。
            n (int): 分区页数，默认为 None (text 模式下无效)。
            verbose (bool): 是否启用详细日志输出，默认为 False。
            kwargs (dict): 传递给 unstructured API 的额外参数字典，默认为空字典。
        """
        self.unstructured_api_url = unstructured_api_url  # 设置 unstructured API URL
        self.unstructured_api_key = unstructured_api_key  # 设置 unstructured API 密钥
        self.start = start  # 设置分区起始页码 (text 模式下无效)
        self.n = n  # 设置分区页数 (text 模式下无效)
        self.headers = {'Content-Type': 'application/json'}  # 设置 HTTP 请求头
        self.file_name = file_name  # 设置文件名
        self.extra_kwargs = kwargs  # 设置额外参数
        super().__init__(file_path)  # 调用父类 BasePDFLoader 的初始化方法

    def load(self) -> List[Document]:
        """加载文件并返回 Document 列表.
        调用 get_text_metadata 方法获取文件内容和元数据，并创建 Langchain Document 对象。
        Returns:
            List[Document]: 包含加载的文件内容的 Document 对象列表，通常只包含一个 Document 对象。
        """
        page_content, metadata = self.get_text_metadata()  # 获取文件内容和元数据
        doc = Document(page_content=page_content, metadata=metadata)  # 创建 Langchain Document 对象
        return [doc]  # 返回包含 Document 对象的列表

    def get_text_metadata(self):
        """获取文件文本内容和元数据.
        调用 unstructured API 的 text 模式解析文件，如果解析失败，则尝试转换为 PDF 后再进行 partition 模式解析。
        Returns:
            Tuple[str, Dict]: 文件文本内容和元数据字典。
        Raises:
            Exception: 如果文件解析失败，抛出异常。
        """
        b64_data = base64.b64encode(open(self.file_path, 'rb').read()).decode()  # 将文件内容 base64 编码
        payload = dict(filename=os.path.basename(self.file_name), b64_data=[b64_data], mode='text')  # 构建 API 请求 payload，指定 text 模式
        payload.update({'start': self.start, 'n': self.n})  # 合并 start 和 n 参数 (text 模式下无效)
        payload.update(self.extra_kwargs)  # 合并额外参数
        resp = requests.post(self.unstructured_api_url, headers=self.headers, json=payload)  # 发送 POST 请求到 unstructured API，使用 text 模式
        # 说明文件解析成功
        if resp.status_code == 200 and resp.json().get('status_code') == 200:  # 如果 API 响应状态码为 200 且返回状态码也为 200
            res = resp.json()  # 解析 JSON 响应
            return res['text'], {'source': self.file_name}  # 返回 text 字段内容和包含 source 字段的元数据字典
        # 说明文件解析失败，pdf文件直接返回报错
        if self.file_name.endswith('.pdf'):  # 如果文件名以 .pdf 结尾，表示原始文件是 PDF 文件
            raise Exception(f'文件文本提取 {os.path.basename(self.file_name)} 失败 resp={resp.text}')  # 抛出异常，提示 PDF 文件文本提取失败
        # 非 pdf 文件，先将文件转为 pdf 格式，然后再执行 partition 模式解析文档
        # 把文件转为 pdf
        resp = requests.post(self.unstructured_api_url, headers=self.headers,
            json={'filename': os.path.basename(self.file_name), 'b64_data': [b64_data], 'mode': 'topdf', })  # 发送 POST 请求到 unstructured API，使用 topdf 模式将文件转换为 PDF
        if resp.status_code != 200 or resp.json().get('status_code') != 200:  # 如果 API 响应状态码不是 200 或返回状态码不是 200，表示 PDF 转换失败
            raise Exception(f'文件转换为 PDF {os.path.basename(self.file_name)} 失败 resp={resp.text}')  # 抛出异常，提示文件转换为 PDF 失败
        # 解析 pdf 文件
        payload['mode'] = 'partition'  # 设置 payload 的 mode 为 partition，使用分区模式解析 PDF
        payload['b64_data'] = [resp.json()['b64_pdf']]  # 使用转换后的 PDF 文件的 base64 编码数据
        payload['filename'] = os.path.basename(self.file_name) + '.pdf'  # 更新 payload 的 filename 为 PDF 文件名
        resp = requests.post(self.unstructured_api_url, headers=self.headers, json=payload)  # 再次发送 POST 请求到 unstructured API，使用 partition 模式解析 PDF
        if resp.status_code != 200 or resp.json().get('status_code') != 200:  # 如果 API 响应状态码不是 200 或返回状态码不是 200，表示 PDF 分区解析失败
            raise Exception(f'文件分区 {os.path.basename(self.file_name)} 失败 resp={resp.text}')  # 抛出异常，提示 PDF 文件分区解析失败
        res = resp.json()  # 解析 JSON 响应
        partitions = res['partitions']  # 获取分区数据
        if not partitions:  # 如果分区数据为空
            raise Exception(f'文件分区结果为空 {os.path.basename(self.file_name)} resp={resp.text}')  # 抛出异常，提示文件分区结果为空
        # 拼接结果为文本
        content, _ = merge_partitions(partitions)  # 合并分区数据，获取文本内容 (忽略元数据)
        return content, {'source': self.file_name}  # 返回合并后的文本内容和包含 source 字段的元数据字典



from .custom_kv import CustomKVLoader
from .elem_pdf import PDFWithSemanticLoader
from .elem_unstrcutured_loader import ElemUnstructuredLoader, ElemUnstructuredLoaderV0
from .universal_kv import UniversalKVLoader

__all__ = [
    'PDFWithSemanticLoader',
    'ElemUnstructuredLoader',
    'ElemUnstructuredLoaderV0',
    'UniversalKVLoader',
    'CustomKVLoader',
]



import os
from typing import Callable, Dict, Optional, Union

import openai
from langchain_community.utilities.dalle_image_generator import DallEAPIWrapper
from langchain_core.pydantic_v1 import Field, SecretStr, root_validator
from langchain_core.utils import convert_to_secret_str, get_from_dict_or_env


class AzureDallEWrapper(DallEAPIWrapper):
    """`Azure OpenAI` Embeddings API.

    To use, you should have the
    environment variable ``AZURE_OPENAI_API_KEY`` set with your API key or pass it
    as a named parameter to the constructor.

    Example:
        .. code-block:: python

            from langchain_openai import AzureOpenAIEmbeddings

            openai = AzureOpenAIEmbeddings(model="text-embedding-3-large")
    """

    azure_endpoint: Union[str, None] = None
    """Your Azure endpoint, including the resource.

        Automatically inferred from env var `AZURE_OPENAI_ENDPOINT` if not provided.

        Example: `https://example-resource.azure.openai.com/`
    """
    deployment: Optional[str] = Field(default=None, alias='azure_deployment')
    """A model deployment.

        If given sets the base client URL to include `/deployments/{azure_deployment}`.
        Note: this means you won't be able to use non-deployment endpoints.
    """
    openai_api_key: Optional[SecretStr] = Field(default=None, alias='api_key')
    """Automatically inferred from env var `AZURE_OPENAI_API_KEY` if not provided."""
    azure_ad_token: Optional[SecretStr] = None
    """Your Azure Active Directory token.

        Automatically inferred from env var `AZURE_OPENAI_AD_TOKEN` if not provided.

        For more:
        https://www.microsoft.com/en-us/security/business/identity-access/microsoft-entra-id.
    """
    azure_ad_token_provider: Union[Callable[[], str], None] = None
    """A function that returns an Azure Active Directory token.

        Will be invoked on every request.
    """
    openai_api_version: Optional[str] = Field(default=None, alias='api_version')
    """Automatically inferred from env var `OPENAI_API_VERSION` if not provided."""
    validate_base_url: bool = True
    chunk_size: int = 2048
    """Maximum number of texts to embed in each batch"""

    @root_validator()
    def validate_environment(cls, values: Dict) -> Dict:
        """Validate that api key and python package exists in environment."""
        # Check OPENAI_KEY for backwards compatibility.
        # TODO: Remove OPENAI_API_KEY support to avoid possible conflict when using
        # other forms of azure credentials.
        openai_api_key = (values['openai_api_key'] or os.getenv('AZURE_OPENAI_API_KEY')
                          or os.getenv('OPENAI_API_KEY'))
        values['openai_api_key'] = (convert_to_secret_str(openai_api_key)
                                    if openai_api_key else None)
        values['openai_api_base'] = (values['openai_api_base'] if 'openai_api_base' in values else
                                     os.getenv('OPENAI_API_BASE'))
        values['openai_api_version'] = values['openai_api_version'] or os.getenv(
            'OPENAI_API_VERSION', default='2023-05-15')
        values['openai_api_type'] = get_from_dict_or_env(values,
                                                         'openai_api_type',
                                                         'OPENAI_API_TYPE',
                                                         default='azure')
        values['openai_organization'] = (values['openai_organization']
                                         or os.getenv('OPENAI_ORG_ID')
                                         or os.getenv('OPENAI_ORGANIZATION'))
        values['openai_proxy'] = get_from_dict_or_env(values,
                                                      'openai_proxy',
                                                      'OPENAI_PROXY',
                                                      default='')
        values['azure_endpoint'] = values['azure_endpoint'] or os.getenv('AZURE_OPENAI_ENDPOINT')
        azure_ad_token = values['azure_ad_token'] or os.getenv('AZURE_OPENAI_AD_TOKEN')
        values['azure_ad_token'] = (convert_to_secret_str(azure_ad_token)
                                    if azure_ad_token else None)
        # For backwards compatibility. Before openai v1, no distinction was made
        # between azure_endpoint and base_url (openai_api_base).
        openai_api_base = values['openai_api_base']
        if openai_api_base and values['validate_base_url']:
            if '/openai' not in openai_api_base:
                values['openai_api_base'] += '/openai'
                raise ValueError('As of openai>=1.0.0, Azure endpoints should be specified via '
                                 'the `azure_endpoint` param not `openai_api_base` '
                                 '(or alias `base_url`). ')
            if values['deployment']:
                raise ValueError('As of openai>=1.0.0, if `deployment` (or alias '
                                 '`azure_deployment`) is specified then '
                                 '`openai_api_base` (or alias `base_url`) should not be. '
                                 'Instead use `deployment` (or alias `azure_deployment`) '
                                 'and `azure_endpoint`.')
        client_params = {
            'api_version':
            values['openai_api_version'],
            'azure_endpoint':
            values['azure_endpoint'],
            'azure_deployment':
            values['deployment'],
            'api_key':
            (values['openai_api_key'].get_secret_value() if values['openai_api_key'] else None),
            'azure_ad_token':
            (values['azure_ad_token'].get_secret_value() if values['azure_ad_token'] else None),
            'azure_ad_token_provider':
            values['azure_ad_token_provider'],
            'organization':
            values['openai_organization'],
            'base_url':
            values['openai_api_base'],
            'timeout':
            values['request_timeout'],
            'max_retries':
            values['max_retries'],
            'default_headers':
            values['default_headers'],
            'default_query':
            values['default_query'],
        }
        if not values.get('client'):
            values['client'] = openai.AzureOpenAI(**client_params).images
        if not values.get('async_client'):
            values['async_client'] = openai.AsyncAzureOpenAI(**client_params).images
        return values

    @property
    def _llm_type(self) -> str:
        return 'azure-openai-chat'



from bisheng_langchain.autogen_role.assistant import AutoGenAssistant
from bisheng_langchain.autogen_role.custom import AutoGenCustomRole
from bisheng_langchain.autogen_role.groupchat_manager import AutoGenGroupChatManager
from bisheng_langchain.autogen_role.user import AutoGenCoder, AutoGenUser

__all__ = ['AutoGenAssistant', 'AutoGenGroupChatManager',
           'AutoGenUser', 'AutoGenCoder',
           'AutoGenCustomRole']
